"""Base client for ANY.RUN APIs."""

from typing import Any, Dict, Optional, Union, cast

import httpx
from loguru import logger
from pydantic import HttpUrl
from redis.asyncio import Redis as AsyncRedis

from .config import BaseConfig
from .constants import (
    API_BASE_URL,
    DEFAULT_TIMEOUT,
    DEFAULT_USER_AGENT,
    HEADER_API_KEY,
    HEADER_USER_AGENT,
)
from .exceptions import APIError, AuthenticationError, NotFoundError, RateLimitError, ServerError
from .sandbox import create_sandbox_client
from .ti_lookup import TILookupClient
from .ti_yara import TIYaraClient
from .types import CacheBackend, RetryStrategy
from .utils import Cache, RateLimiter, retry
from .utils.cache import CacheBackend as CacheBackendType
from .utils.cache import MemoryCache


class BaseClient:
    """Base client for ANY.RUN APIs."""

    def __init__(self, config: BaseConfig, client: Optional[httpx.AsyncClient] = None) -> None:
        """Initialize base client.

        Args:
            config: Client configuration
            client: Optional pre-configured HTTP client
        """
        self.config = config
        self._client = client
        self._cache: Optional[Cache] = None
        self._rate_limiter: Optional[RateLimiter] = None

    async def _ensure_client(self) -> httpx.AsyncClient:
        """Ensure HTTP client is initialized.

        Returns:
            httpx.AsyncClient: HTTP client instance
        """
        if self._client is None:
            self._client = httpx.AsyncClient(
                base_url=str(self.config.base_url),
                timeout=self.config.timeout,
                headers=self._get_default_headers(),
                verify=self.config.verify_ssl,
                follow_redirects=True,
            )
        return self._client

    def _get_default_headers(self) -> Dict[str, str]:
        """Get default HTTP headers.

        Returns:
            Dict[str, str]: Default headers
        """
        headers = {
            HEADER_USER_AGENT: self.config.user_agent or DEFAULT_USER_AGENT,
            HEADER_API_KEY: self.config.api_key,
            "Accept": "application/json",
        }
        headers.update(self.config.headers)
        return headers

    @property
    def cache(self) -> Cache:
        """Get cache instance.

        Returns:
            Cache: Cache instance
        """
        if self._cache is None:
            backend: Union[CacheBackendType, AsyncRedis[Any], None] = None
            if self.config.cache_backend == CacheBackend.REDIS:
                backend = AsyncRedis()
            elif self.config.cache_backend == CacheBackend.MEMORY:
                backend = MemoryCache()

            self._cache = Cache(
                backend=backend,
                enabled=self.config.cache_enabled,
                prefix=f"anyrun:{self.config.api_version}:",
                default_ttl=self.config.cache_ttl,
            )
        return self._cache

    @property
    def rate_limiter(self) -> RateLimiter:
        """Get rate limiter instance.

        Returns:
            RateLimiter: Rate limiter instance
        """
        if self._rate_limiter is None:
            self._rate_limiter = RateLimiter(
                rate=1.0,  # Will be set per endpoint
                burst=10,
                key=f"anyrun:{self.config.api_version}",
            )
        return self._rate_limiter

    async def _request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        cache_key: Optional[str] = None,
        cache_ttl: Optional[int] = None,
        rate_limit: Optional[float] = None,
    ) -> Dict[str, Any]:
        """Make HTTP request without retry.

        Args:
            method: HTTP method
            url: Request URL
            params: Query parameters
            json: JSON body
            data: Form data
            files: Files to upload
            headers: Additional headers
            cache_key: Cache key
            cache_ttl: Cache TTL in seconds
            rate_limit: Rate limit in requests per second

        Returns:
            Dict[str, Any]: Response data

        Raises:
            APIError: If request fails
        """
        # Check cache
        if cache_key and method.upper() == "GET":
            cached = await self.cache.get(cache_key)
            if cached is not None:
                return cast(Dict[str, Any], cached)

        # Apply rate limiting
        if rate_limit and self.config.rate_limit_enabled:
            self.rate_limiter.rate = rate_limit
            if self.config.retry_strategy == RetryStrategy.EXPONENTIAL:
                await self.rate_limiter.acquire()
            elif not await self.rate_limiter.check():
                raise RateLimitError(
                    message="Rate limit exceeded",
                    retry_after=int(1.0 / rate_limit),
                )

        # Make request
        client = await self._ensure_client()
        try:
            response = await client.request(
                method=method,
                url=url,
                params=params,
                json=json,
                data=data,
                files=files,
                headers=headers,
            )
            response.raise_for_status()
            response_data: Dict[str, Any] = response.json()

            # Cache successful GET response
            if cache_key and method.upper() == "GET":
                await self.cache.set(cache_key, response_data, ttl=cache_ttl)

            return response_data

        except httpx.HTTPStatusError as e:
            error_msg = str(e)
            try:
                error_data = e.response.json()
                if isinstance(error_data, dict):
                    error_msg = error_data.get("message", str(e))
            except Exception:
                # Failed to parse error response, use original error message
                logger.debug("Failed to parse error response JSON")

            if e.response.status_code == 401:
                raise AuthenticationError(error_msg) from e
            elif e.response.status_code == 404:
                raise NotFoundError(error_msg) from e
            elif e.response.status_code == 429:
                raise RateLimitError(
                    message=error_msg,
                    retry_after=int(float(e.response.headers.get("Retry-After", "60"))),
                )
            elif e.response.status_code >= 500:
                raise ServerError(error_msg)
            else:
                raise APIError(error_msg) from e

        except httpx.RequestError as e:
            raise APIError(f"Request failed: {str(e)}")

    async def request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        cache_key: Optional[str] = None,
        cache_ttl: Optional[int] = None,
        rate_limit: Optional[float] = None,
    ) -> Dict[str, Any]:
        """Make HTTP request.

        Args:
            method: HTTP method
            url: Request URL
            params: Query parameters
            json: JSON body
            data: Form data
            files: Files to upload
            headers: Additional headers
            cache_key: Cache key
            cache_ttl: Cache TTL in seconds
            rate_limit: Rate limit in requests per second

        Returns:
            Dict[str, Any]: Response data

        Raises:
            APIError: If request fails
        """

        async def _request_with_config(*args: Any, **kwargs: Any) -> Dict[str, Any]:
            return await self._request(*args, **kwargs)

        _request_with_config.config = self.config  # type: ignore

        retry_decorator = retry(
            max_attempts=self.config.retry_max_attempts,
            delay=self.config.retry_initial_delay,
            max_delay=self.config.retry_max_delay,
            exponential=self.config.retry_strategy == RetryStrategy.EXPONENTIAL,
            jitter=True,
        )

        return await retry_decorator(_request_with_config)(
            method=method,
            url=url,
            params=params,
            json=json,
            data=data,
            files=files,
            headers=headers,
            cache_key=cache_key,
            cache_ttl=cache_ttl,
            rate_limit=rate_limit,
        )

    async def close(self) -> None:
        """Close HTTP client."""
        if self._client is not None:
            await self._client.aclose()
            self._client = None

    async def __aenter__(self) -> "BaseClient":
        """Enter async context."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Exit async context."""
        await self.close()

    async def test_request(self, url: str) -> Dict[str, Any]:
        """Make test request.

        Args:
            url: Request URL

        Returns:
            Dict[str, Any]: Response data

        Raises:
            APIError: If request fails
        """
        return await self._request("GET", url)


class AnyRunClient:
    """Main client for ANY.RUN APIs."""

    def __init__(
        self,
        api_key: str,
        sandbox_version: str = "v1",
        ti_lookup_version: str = "v1",
        ti_yara_version: str = "v1",
        **kwargs: Any,
    ) -> None:
        """Initialize ANY.RUN client.

        Args:
            api_key: ANY.RUN API key
            sandbox_version: Sandbox API version
            ti_lookup_version: TI Lookup API version
            ti_yara_version: TI YARA API version
            **kwargs: Additional configuration options
        """
        self.config = BaseConfig(
            api_key=api_key,
            base_url=HttpUrl(API_BASE_URL),
            timeout=kwargs.get("timeout", DEFAULT_TIMEOUT),
            user_agent=kwargs.get("user_agent", DEFAULT_USER_AGENT),
            verify_ssl=kwargs.get("verify_ssl", True),
            headers=kwargs.get("headers", {}),
            cache_enabled=kwargs.get("cache_enabled", True),
            cache_ttl=kwargs.get("cache_ttl", 300),
            cache_backend=kwargs.get("cache_backend", CacheBackend.MEMORY),
            rate_limit_enabled=kwargs.get("rate_limit_enabled", True),
            retry_strategy=kwargs.get("retry_strategy", RetryStrategy.EXPONENTIAL),
        )

        self._base_client = BaseClient(self.config)

        # Initialize API clients
        self.sandbox = create_sandbox_client(
            api_key=api_key,
            version=sandbox_version,
            base_url=self.config.base_url,
            timeout=int(self.config.timeout),
            verify_ssl=self.config.verify_ssl,
            user_agent=self.config.user_agent,
            headers=self.config.headers,
        )
        self.ti_lookup = TILookupClient(
            api_key=api_key,
            version=ti_lookup_version,
            cache_enabled=self.config.cache_enabled,
            timeout=int(self.config.timeout),
        )
        self.ti_yara = TIYaraClient(
            api_key=api_key,
            version=ti_yara_version,
            cache_enabled=self.config.cache_enabled,
            timeout=int(self.config.timeout),
        )

    @property
    def cache(self) -> Cache:
        """Get cache instance.

        Returns:
            Cache: Cache instance
        """
        return self._base_client.cache

    @property
    def rate_limiter(self) -> RateLimiter:
        """Get rate limiter instance.

        Returns:
            RateLimiter: Rate limiter instance
        """
        return self._base_client.rate_limiter

    async def test_request(self, url: str) -> Dict[str, Any]:
        """Make test request.

        Args:
            url: Request URL

        Returns:
            Dict[str, Any]: Response data

        Raises:
            APIError: If request fails
        """
        return await self._base_client.request("GET", url)

    async def close(self) -> None:
        """Close all clients."""
        await self.sandbox.close()
        await self.ti_lookup.close()
        await self.ti_yara.close()
        await self._base_client.close()

    async def __aenter__(self) -> "AnyRunClient":
        """Enter async context."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Exit async context."""
        await self.close()



"""Sandbox API client."""

from typing import Dict, Optional

from pydantic import HttpUrl

from .base import BaseSandboxClient
from .v1.client import SandboxClientV1


def create_sandbox_client(
    api_key: str,
    version: str = "v1",
    base_url: Optional[HttpUrl] = None,
    timeout: Optional[int] = None,
    verify_ssl: bool = True,
    user_agent: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
) -> BaseSandboxClient:
    """Create sandbox client instance.

    Args:
        api_key: ANY.RUN API key
        version: API version
        base_url: Base URL for API requests
        timeout: Request timeout in seconds
        verify_ssl: Verify SSL certificates
        user_agent: User agent string
        headers: Additional headers

    Returns:
        BaseSandboxClient: Sandbox client instance

    Raises:
        ValueError: If version is not supported
    """
    if version == "v1":
        return SandboxClientV1(
            api_key=api_key,
            base_url=base_url,
            timeout=timeout,
            verify_ssl=verify_ssl,
            user_agent=user_agent,
            headers=headers,
        )
    raise ValueError(f"Unsupported version: {version}")


__all__ = ["create_sandbox_client", "BaseSandboxClient"]



"""Utility functions and classes for ANY.RUN Tools."""

from .cache import Cache, CacheBackend, NoCache, RedisCache
from .rate_limit import RateLimiter, RateLimitError
from .retry import RetryConfig, RetryError, retry
from .validation import ValidationError, validate_api_key, validate_file_size

__all__ = [
    "Cache",
    "CacheBackend",
    "RedisCache",
    "NoCache",
    "retry",
    "RetryConfig",
    "RetryError",
    "RateLimiter",
    "RateLimitError",
    "validate_api_key",
    "validate_file_size",
    "ValidationError",
]



"""Schemas for Sandbox API data validation."""

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field, HttpUrl, model_validator


class ObjectType(str, Enum):
    """Analysis object type."""

    FILE = "file"
    URL = "url"
    DOWNLOAD = "download"
    RERUN = "rerun"


class OSType(str, Enum):
    """Operating system type."""

    WINDOWS = "windows"
    LINUX = "linux"


class BitnessType(str, Enum):
    """Operation System bitness type.

    For Linux: only 64
    For Win11: only 64
    For Win7/10: 32 or 64
    """

    X32 = "32"
    X64 = "64"


class WindowsVersion(str, Enum):
    """Windows version."""

    WIN7 = "7"
    WIN10 = "10"
    WIN11 = "11"


class LinuxVersion(str, Enum):
    """Linux version."""

    UBUNTU_22_04_2 = "22.04.2"


class EnvType(str, Enum):
    """Environment type.

    For Windows: clean, office, complete
    For Linux: only office
    """

    CLEAN = "clean"
    OFFICE = "office"
    COMPLETE = "complete"


class Browser(str, Enum):
    """Browser type.

    Default value for Windows: Microsoft Edge
    Allowed values for Windows: Google Chrome, Mozilla Firefox, Internet Explorer, Microsoft Edge
    Default value for Linux: Google Chrome
    Allowed values for Linux: Mozilla Firefox, Google Chrome
    """

    CHROME = "Google Chrome"
    FIREFOX = "Mozilla Firefox"
    IE = "Internet Explorer"
    EDGE = "Microsoft Edge"


class PrivacyType(str, Enum):
    """Privacy type."""

    PUBLIC = "public"
    BYLINK = "bylink"
    OWNER = "owner"
    BYTEAM = "byteam"


class StartFolder(str, Enum):
    """Start folder.

    Linux: desktop, downloads, home, temp
    Windows: appdata, desktop, downloads, home, root, temp, windows
    """

    DESKTOP = "desktop"
    DOWNLOADS = "downloads"
    HOME = "home"
    TEMP = "temp"
    APPDATA = "appdata"  # Windows only
    ROOT = "root"  # Windows only
    WINDOWS = "windows"  # Windows only


class AnalysisRequest(BaseModel):
    """Analysis request model."""

    # Object options
    obj_type: ObjectType = Field(description="Object type")
    obj_url: Optional[HttpUrl] = Field(None, description="Object URL")
    obj_hash: Optional[str] = Field(None, description="Object hash")
    obj_content: Optional[str] = Field(None, description="Object content")
    obj_filename: Optional[str] = Field(None, description="Object filename")

    # Environment options
    env_os: OSType = Field(description="Operating system type")
    env_bitness: BitnessType = Field(description="Operating system bitness")
    env_version: Optional[str] = Field(None, description="Operating system version")
    env_type: EnvType = Field(description="Environment type")
    env_browser: Optional[Browser] = Field(None, description="Browser type")

    # Object extension options
    obj_ext_runasadmin: Optional[bool] = Field(None, description="Run as admin")
    obj_ext_runwithargs: Optional[str] = Field(None, description="Run with arguments")
    obj_ext_runwithfile: Optional[str] = Field(None, description="Run with file")
    obj_ext_runwithurl: Optional[HttpUrl] = Field(None, description="Run with URL")
    obj_ext_runwithcontent: Optional[str] = Field(None, description="Run with content")
    obj_ext_runwithfilename: Optional[str] = Field(
        None,
        description="Run with filename",
        min_length=2,
        max_length=256,
    )
    obj_ext_elevateprompt: Optional[bool] = Field(
        None,
        description="Windows only",
    )
    obj_force_elevation: Optional[bool] = Field(
        None,
        description="Windows only",
    )
    auto_confirm_uac: Optional[bool] = Field(
        default=True,
        description="Windows only",
    )
    run_as_root: Optional[bool] = Field(
        default=False,
        description="Linux only",
    )
    obj_ext_extension: Optional[bool] = None
    obj_ext_startfolder: Optional[StartFolder] = Field(
        default=StartFolder.TEMP, description="Start object folder"
    )

    # Network options
    opt_network_connect: Optional[bool] = Field(
        default=True, description="Network connection state"
    )
    opt_network_fakenet: Optional[bool] = Field(default=False, description="FakeNet feature status")
    opt_network_tor: Optional[bool] = Field(default=False, description="TOR using")
    opt_network_mitm: Optional[bool] = Field(default=False, description="HTTPS MITM proxy option")

    # Privacy options
    opt_privacy_type: Optional[PrivacyType] = Field(
        default=PrivacyType.BYLINK, description="Privacy settings"
    )
    opt_privacy_hidesource: Optional[bool] = Field(
        default=False, description="Option for hiding source URL"
    )

    # Advanced options
    opt_chatgpt: Optional[bool] = None

    @model_validator(mode="after")  # type: ignore[misc]
    def validate_request(self) -> "AnalysisRequest":
        """Validate request.

        Returns:
            AnalysisRequest: Validated request

        Raises:
            ValueError: If request is invalid
        """
        # Validate object type
        if self.obj_type == ObjectType.FILE:
            if not self.obj_content:
                raise ValueError("Object content is required for file analysis")
            if not self.obj_filename:
                raise ValueError("Object filename is required for file analysis")
        elif self.obj_type == ObjectType.URL:
            if not self.obj_url:
                raise ValueError("Object URL is required for URL analysis")
        elif self.obj_type == ObjectType.DOWNLOAD:
            if not self.obj_url:
                raise ValueError("Object URL is required for download analysis")
        elif self.obj_type == ObjectType.RERUN:
            if not self.obj_hash:
                raise ValueError("Object hash is required for rerun analysis")

        # Validate environment
        if self.env_os == OSType.WINDOWS:
            if self.env_version not in WindowsVersion._value2member_map_:
                raise ValueError(
                    "Invalid Windows version. Supported: "
                    f"{list(WindowsVersion._value2member_map_.keys())}"
                )
            if self.env_version == WindowsVersion.WIN11 and self.env_bitness != BitnessType.X64:
                raise ValueError("Windows 11 supports only 64-bit")
            if self.env_type == EnvType.OFFICE and self.env_version == WindowsVersion.WIN11:
                raise ValueError("Windows 11 does not support office environment")
            if self.run_as_root:
                raise ValueError("Run as root is not supported for Windows")
        elif self.env_os == OSType.LINUX:
            if self.env_version not in LinuxVersion._value2member_map_:
                raise ValueError(
                    "Invalid Linux version. Supported: "
                    f"{list(LinuxVersion._value2member_map_.keys())}"
                )
            if self.env_bitness != BitnessType.X64:
                raise ValueError("Linux supports only 64-bit")
            if self.env_type != EnvType.OFFICE:
                raise ValueError("Linux supports only office environment")
            if self.obj_ext_elevateprompt or self.obj_force_elevation or self.auto_confirm_uac:
                raise ValueError("UAC options are not supported for Linux")

        # Validate browser
        if self.env_browser:
            if self.env_os == OSType.WINDOWS:
                if self.env_browser not in [
                    Browser.CHROME,
                    Browser.FIREFOX,
                    Browser.IE,
                    Browser.EDGE,
                ]:
                    raise ValueError(
                        "Invalid browser for Windows. Supported: "
                        "Google Chrome, Mozilla Firefox, Internet Explorer, Microsoft Edge"
                    )
            elif self.env_os == OSType.LINUX:
                if self.env_browser not in [Browser.CHROME, Browser.FIREFOX]:
                    raise ValueError(
                        "Invalid browser for Linux. Supported: " "Google Chrome, Mozilla Firefox"
                    )

        # Validate start folder
        if self.obj_ext_startfolder:
            if self.env_os == OSType.WINDOWS:
                if self.obj_ext_startfolder not in [
                    StartFolder.APPDATA,
                    StartFolder.DESKTOP,
                    StartFolder.DOWNLOADS,
                    StartFolder.HOME,
                    StartFolder.ROOT,
                    StartFolder.TEMP,
                    StartFolder.WINDOWS,
                ]:
                    raise ValueError(
                        "Invalid start folder for Windows. Supported: "
                        "appdata, desktop, downloads, home, root, temp, windows"
                    )
            elif self.env_os == OSType.LINUX:
                if self.obj_ext_startfolder not in [
                    StartFolder.DESKTOP,
                    StartFolder.DOWNLOADS,
                    StartFolder.HOME,
                    StartFolder.TEMP,
                ]:
                    raise ValueError(
                        "Invalid start folder for Linux. Supported: desktop, downloads, home, temp"
                    )

        return self


class AnalysisData(BaseModel):
    """Analysis data model."""

    task_id: str
    status: str


class AnalysisResponse(BaseModel):
    """Analysis response."""

    error: bool
    data: AnalysisData


class AnalysisListRequest(BaseModel):
    """Analysis list request parameters."""

    team: bool = Field(default=False, description="Get team history instead of personal")
    skip: int = Field(default=0, ge=0, description="Number of items to skip")
    limit: int = Field(default=25, ge=1, le=100, description="Number of items per page")


class AnalysisListItem(BaseModel):
    """Analysis list item model."""

    task_id: str
    status: str


class AnalysisListData(BaseModel):
    """Analysis list data model."""

    items: List[AnalysisListItem]
    total: int


class AnalysisListResponse(BaseModel):
    """Analysis list response."""

    error: bool
    data: AnalysisListData


class WindowsEnvironment(BaseModel):
    """Windows environment model."""

    versions: List[str]
    bitness: List[str]


class LinuxEnvironment(BaseModel):
    """Linux environment model."""

    versions: List[str]
    bitness: List[str]


class EnvironmentData(BaseModel):
    """Environment data model."""

    windows: WindowsEnvironment
    linux: LinuxEnvironment


class EnvironmentResponse(BaseModel):
    """Environment information response."""

    error: bool
    data: EnvironmentData


class AddTimeResponse(BaseModel):
    """Response model for add time request."""

    error: bool = Field(description="Error flag")
    message: str = Field(description="Response message")


class StopAnalysisResponse(BaseModel):
    """Stop analysis response."""

    error: bool
    data: dict


class DeleteAnalysisResponse(BaseModel):
    """Delete analysis response."""

    error: bool
    data: dict



"""Common constants for ANY.RUN Tools."""

from enum import Enum
from typing import Dict, Final, Tuple

from ._version import __version__

# API URLs
API_BASE_URL: Final[str] = "https://api.any.run"
API_DOCS_URL: Final[str] = "https://api.any.run/docs"


# API versions
class APIVersion(str, Enum):
    """API version."""

    V1 = "v1"
    V2 = "v2"
    LATEST = V1


# HTTP headers
HEADER_API_KEY: Final[str] = "Authorization"
HEADER_USER_AGENT: Final[str] = "User-Agent"
DEFAULT_USER_AGENT: Final[str] = f"anyrun-tools/{__version__}"

# Request limits
MAX_FILE_SIZE: Final[int] = 100 * 1024 * 1024  # 100MB
MAX_URL_LENGTH: Final[int] = 2048
MAX_BATCH_SIZE: Final[int] = 100

# Rate limits (requests per minute)
RATE_LIMITS: Final[Dict[str, Dict[str, int]]] = {
    "sandbox": {"analyze": 10, "list": 60, "status": 120, "download": 30},
    "ti_lookup": {"hash": 60, "url": 60, "ip": 60, "domain": 60},
    "ti_yara": {"match": 30, "list": 60},
}

# Cache TTLs (seconds)
CACHE_TTLS: Final[Dict[str, int]] = {
    "environment": 3600,  # 1 hour
    "user_info": 300,  # 5 minutes
    "analysis": 60,  # 1 minute
    "ti_lookup": 1800,  # 30 minutes
    "ti_yara": 1800,  # 30 minutes
}

# Retry configuration
DEFAULT_RETRY_COUNT: Final[int] = 3
DEFAULT_RETRY_DELAY: Final[float] = 1.0
MAX_RETRY_DELAY: Final[float] = 10.0

# Webhook events
WEBHOOK_EVENTS: Final[Tuple[str, ...]] = (
    "analysis.started",
    "analysis.finished",
    "analysis.failed",
    "ti_lookup.matched",
    "ti_yara.matched",
)

# File extensions
SUPPORTED_EXTENSIONS: Final[Tuple[str, ...]] = (
    # Executables
    ".exe",
    ".dll",
    ".sys",
    ".scr",
    ".com",
    # Scripts
    ".ps1",
    ".vbs",
    ".js",
    ".jse",
    ".hta",
    # Documents
    ".doc",
    ".docx",
    ".xls",
    ".xlsx",
    ".pdf",
    # Archives
    ".zip",
    ".7z",
    ".rar",
    ".tar",
    ".gz",
)

# Environment defaults
DEFAULT_TIMEOUT: Final[float] = 30.0  # seconds
DEFAULT_CACHE_TTL: Final[int] = 300  # seconds
DEFAULT_CACHE_PREFIX: Final[str] = "anyrun:"

# Rate limit settings
DEFAULT_RATE_LIMIT: Final[int] = 10  # requests per second
DEFAULT_RATE_LIMIT_WINDOW: Final[float] = 1.0  # seconds

# HTTP status codes
HTTP_STATUS_CODES: Final[Dict[str, int]] = {
    "OK": 200,
    "BAD_REQUEST": 400,
    "UNAUTHORIZED": 401,
    "NOT_FOUND": 404,
    "TOO_MANY_REQUESTS": 429,
    "INTERNAL_SERVER_ERROR": 500,
}

# HTTP methods
HTTP_METHODS: Final[Tuple[str, ...]] = (
    "GET",
    "POST",
    "PUT",
    "DELETE",
    "HEAD",
    "OPTIONS",
)

# File extensions
FILE_EXTENSIONS: Final[Tuple[str, ...]] = (
    ".exe",
    ".dll",
    ".sys",
    ".scr",
    ".cpl",
    ".ocx",
    ".ax",
    ".bin",
    ".cmd",
    ".bat",
    ".com",
    ".js",
    ".jse",
    ".vbs",
    ".vbe",
    ".wsf",
    ".wsh",
    ".ps1",
    ".msi",
    ".doc",
    ".docx",
    ".xls",
    ".xlsx",
    ".ppt",
    ".pptx",
    ".pdf",
    ".rtf",
    ".html",
    ".htm",
    ".hta",
    ".swf",
    ".jar",
    ".class",
    ".zip",
    ".rar",
    ".7z",
    ".tar",
    ".gz",
    ".bz2",
    ".iso",
    ".img",
    ".vhd",
    ".vmdk",
    ".eml",
    ".msg",
    ".url",
    ".lnk",
    ".reg",
    ".inf",
    ".scf",
    ".job",
    ".tmp",
    ".dat",
    ".log",
    ".txt",
    ".cfg",
    ".ini",
    ".db",
    ".sqlite",
    ".mdb",
    ".accdb",
    ".pst",
    ".ost",
    ".bak",
    ".old",
    ".new",
)



"""Validation utilities for ANY.RUN Tools."""

import os
import re
from typing import Any, Type, TypeVar, Union, cast

from loguru import logger
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class ValidationError(Exception):
    """Validation error."""


def validate_api_key(api_key: str) -> None:
    """Validate ANY.RUN API key format.

    Args:
        api_key: API key to validate

    Raises:
        ValidationError: If API key is invalid
    """
    if not api_key:
        raise ValidationError("API key is required")

    # API key format: 32 hex characters
    if not re.match(r"^[a-f0-9]{32}$", api_key.lower()):
        raise ValidationError("Invalid API key format. Expected 32 hexadecimal characters.")


def validate_file_size(
    file_path: Union[str, bytes], max_size: int = 100 * 1024 * 1024  # 100MB
) -> None:
    """Validate file size.

    Args:
        file_path: Path to file or file content as bytes
        max_size: Maximum allowed file size in bytes

    Raises:
        ValidationError: If file size exceeds maximum
        FileNotFoundError: If file does not exist
    """
    try:
        if isinstance(file_path, str):
            if not os.path.exists(file_path):
                raise FileNotFoundError(f"File not found: {file_path}")

            size = os.path.getsize(file_path)
        else:
            size = len(file_path)

        if size > max_size:
            raise ValidationError(
                f"File size ({size} bytes) exceeds maximum allowed size " f"({max_size} bytes)"
            )

        if size == 0:
            raise ValidationError("File is empty")

    except Exception as e:
        if isinstance(e, ValidationError):
            raise
        logger.error(f"File validation error: {str(e)}")
        raise ValidationError(f"File validation failed: {str(e)}")


def validate_url(url: str) -> None:
    """Validate URL format.

    Args:
        url: URL to validate

    Raises:
        ValidationError: If URL is invalid
    """
    if not url:
        raise ValidationError("URL is required")

    # Basic URL format validation
    pattern = (
        r"^https?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?|"  # domain
        r"localhost|"  # localhost
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$"  # path
    )
    if not re.match(pattern, url, re.IGNORECASE):
        raise ValidationError("Invalid URL format")


def validate_hash(hash_value: str, hash_type: str) -> None:
    """Validate hash format.

    Args:
        hash_value: Hash value to validate
        hash_type: Hash type (md5, sha1, sha256)

    Raises:
        ValidationError: If hash is invalid
    """
    if not hash_value:
        raise ValidationError("Hash value is required")

    if hash_type not in ("md5", "sha1", "sha256"):
        raise ValidationError(
            f"Invalid hash type: {hash_type}. " "Supported types: md5, sha1, sha256"
        )

    patterns = {
        "md5": r"^[a-f0-9]{32}$",
        "sha1": r"^[a-f0-9]{40}$",
        "sha256": r"^[a-f0-9]{64}$",
    }

    if not re.match(patterns[hash_type], hash_value.lower()):
        raise ValidationError(
            f"Invalid {hash_type} hash format. "
            f"Expected {len(hash_value)} hexadecimal characters."
        )


def validate_model(model_class: Type[T], data: Any) -> T:
    """Validate data against Pydantic model.

    Args:
        model_class: Pydantic model class
        data: Data to validate

    Returns:
        T: Validated model instance

    Raises:
        ValidationError: If validation fails
    """
    try:
        return cast(T, model_class.model_validate(data))
    except Exception as e:
        logger.error(f"Model validation error: {str(e)}")
        raise ValidationError(f"Model validation failed: {str(e)}")



"""Task status update models for Sandbox API v1."""

from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from .common import HashesApiDto, PrivacyType


class TaskScoresDto(BaseModel):
    """Task scores model."""

    specs: Dict[str, bool] = Field(description="Task specifications")
    verdict: Dict[str, Any] = Field(description="Task verdict")


class TaskObjectDto(BaseModel):
    """Task object model."""

    names: Optional[Dict[str, str]] = Field(None, description="Object names")
    hashes: Optional[HashesApiDto] = Field(None, description="Object hashes")
    urls: Optional[Dict[str, str]] = Field(None, description="Object URLs")


class TaskOptionsDto(BaseModel):
    """Task options model."""

    private: PrivacyType = Field(description="Privacy level")
    whitelist: List[str] = Field(default=[], description="Whitelisted items")
    mitm: bool = Field(description="MITM status")
    fakenet: bool = Field(description="FakeNet status")
    openVPN: Optional[str] = Field(None, description="OpenVPN configuration")
    torGeo: Optional[str] = Field(None, description="TOR geography")
    netviator: bool = Field(description="Netviator status")
    netConnected: bool = Field(description="Network connection status")
    network: str = Field(description="Network type")
    logger: str = Field(description="Logger type")
    presentation: bool = Field(description="Presentation mode")
    teamwork: bool = Field(description="Team collaboration")
    reboots: bool = Field(description="System reboots")
    onlyimportant: bool = Field(description="Important events only")
    video: bool = Field(description="Video recording")
    locale: Optional[str] = Field(None, description="System locale")
    residentialProxyGeo: Optional[str] = Field(None, description="Residential proxy geography")
    residentialProxy: bool = Field(description="Residential proxy status")
    autoclickerDebugMode: Optional[bool] = Field(None, description="Autoclicker debug mode")
    autoclicker: bool = Field(description="Autoclicker status")
    chatGPT: Optional[bool] = Field(None, description="ChatGPT integration")


class TaskEnvironmentDto(BaseModel):
    """Task environment model."""

    OS: Dict[str, Any] = Field(description="Operating system information")
    software: List[Dict[str, Any]] = Field(default=[], description="Installed software")


class TaskPublicDto(BaseModel):
    """Task public information model."""

    maxAddedTimeReached: bool = Field(description="Maximum allowed task runtime reached")
    objects: TaskObjectDto = Field(description="Task objects information")
    options: TaskOptionsDto = Field(description="Task options")
    environment: TaskEnvironmentDto = Field(description="Task environment information")


class TaskTimesDto(BaseModel):
    """Task timing information."""

    created: Optional[datetime] = Field(None, description="Task creation time")
    started: Optional[datetime] = Field(None, description="Task start time")
    completed: Optional[datetime] = Field(None, description="Task completion time")
    addedTime: Optional[datetime] = Field(None, description="Time when additional time was added")


class TaskActionsDto(BaseModel):
    """Task actions model."""

    addTime: Optional[bool] = Field(None, description="Can add time")
    stop: Optional[bool] = Field(None, description="Can stop task")
    delete: Optional[bool] = Field(None, description="Can delete task")


class TaskStatusDto(BaseModel):
    """Task status model."""

    id: Optional[str] = Field(None, alias="_id", description="Internal task ID")
    uuid: str = Field(
        pattern="^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
        description="Task UUID",
    )
    status: int = Field(ge=0, le=100, description="Task completion rate (%)")
    remaining: int = Field(ge=0, description="Number of seconds until task completion")
    times: TaskTimesDto = Field(description="Task timing information")
    public: TaskPublicDto = Field(description="Public task information")
    usersTags: List[str] = Field(default=[], description="User defined tags")
    tags: List[str] = Field(default=[], description="System tags")
    scores: TaskScoresDto = Field(description="Task scores")
    actions: TaskActionsDto = Field(description="Task actions")
    threats: List[str] = Field(default=[], description="Detected threats")


class TaskStatusUpdateDto(BaseModel):
    """Task status update model."""

    task: TaskStatusDto = Field(description="Task status information")
    completed: bool = Field(description="Task completion status")
    error: Literal[False] = Field(False, description="Error status")



"""Constants for Sandbox API."""

from typing import Dict, Final, Set

# API endpoints
ENDPOINT_ANALYSIS: Final[str] = "/analysis"
ENDPOINT_ANALYSIS_STATUS: Final[str] = "/analysis/status/{task_id}"
ENDPOINT_ANALYSIS_MONITOR: Final[str] = "/analysis/monitor/{task_id}"
ENDPOINT_ANALYSIS_ADD_TIME: Final[str] = "/analysis/addtime/{task_id}"
ENDPOINT_ANALYSIS_STOP: Final[str] = "/analysis/stop/{task_id}"
ENDPOINT_ANALYSIS_DELETE: Final[str] = "/analysis/delete/{task_id}"
ENDPOINT_ENVIRONMENT: Final[str] = "/environment"
ENDPOINT_USER: Final[str] = "/user"
ENDPOINT_USER_PRESETS: Final[str] = "/user/presets"

# Cache keys
CACHE_KEY_ENVIRONMENT: Final[str] = "sandbox:environment"
CACHE_KEY_USER_INFO: Final[str] = "sandbox:user_info"
CACHE_KEY_USER_PRESETS: Final[str] = "sandbox:user_presets"
CACHE_KEY_ANALYSIS: Final[str] = "sandbox:analysis:{task_id}"

# Default values
DEFAULT_TIMEOUT: Final[int] = 30
DEFAULT_CACHE_TTL: Final[int] = 300
DEFAULT_RATE_LIMIT: Final[float] = 1.0
DEFAULT_USER_AGENT: Final[str] = "anyrun-tools/sandbox"

# File size limits
MAX_FILE_SIZE: Final[int] = 100 * 1024 * 1024  # 100MB
MAX_MEMORY_SIZE: Final[int] = 500 * 1024 * 1024  # 500MB

# Analysis settings
DEFAULT_MONITOR_TIMEOUT: Final[int] = 1800  # 30 minutes
DEFAULT_MONITOR_INTERVAL: Final[float] = 1.0  # 1 second
DEFAULT_DOWNLOAD_TIMEOUT: Final[int] = 300  # 5 minutes

# Environment settings
SUPPORTED_OS_TYPES: Final[Set[str]] = {"windows", "linux"}

SUPPORTED_WINDOWS_VERSIONS: Final[Set[str]] = {"7", "10", "11"}

SUPPORTED_LINUX_VERSIONS: Final[Set[str]] = {"22.04.2"}  # Ubuntu

SUPPORTED_BITNESS: Final[Dict[str, Set[str]]] = {
    "windows": {"32", "64"},
    "linux": {"64"},
}

SUPPORTED_ENV_TYPES: Final[Dict[str, Set[str]]] = {
    "windows": {"clean", "office", "complete"},
    "linux": {"office"},
}

SUPPORTED_BROWSERS: Final[Dict[str, Set[str]]] = {
    "windows": {
        "Google Chrome",
        "Mozilla Firefox",
        "Internet Explorer",
        "Microsoft Edge",
    },
    "linux": {"Google Chrome", "Mozilla Firefox"},
}

DEFAULT_BROWSERS: Final[Dict[str, str]] = {
    "windows": "Microsoft Edge",
    "linux": "Google Chrome",
}

# Start folders
SUPPORTED_START_FOLDERS: Final[Dict[str, Set[str]]] = {
    "windows": {"appdata", "desktop", "downloads", "home", "root", "temp", "windows"},
    "linux": {"desktop", "downloads", "home", "temp"},
}

# Privacy types
PRIVACY_TYPES: Final[Set[str]] = {"public", "bylink", "owner", "byteam"}

# Analysis status codes
STATUS_QUEUED: Final[str] = "queued"
STATUS_RUNNING: Final[str] = "running"
STATUS_COMPLETED: Final[str] = "completed"
STATUS_FAILED: Final[str] = "failed"
STATUS_STOPPED: Final[str] = "stopped"
STATUS_DELETED: Final[str] = "deleted"

# Analysis events
EVENT_STARTED: Final[str] = "analysis.started"
EVENT_FINISHED: Final[str] = "analysis.finished"
EVENT_FAILED: Final[str] = "analysis.failed"
EVENT_STOPPED: Final[str] = "analysis.stopped"
EVENT_DELETED: Final[str] = "analysis.deleted"

# File extensions
SUPPORTED_EXTENSIONS: Final[Dict[str, Set[str]]] = {
    "executables": {".exe", ".dll", ".sys", ".scr", ".com"},
    "scripts": {".ps1", ".vbs", ".js", ".jse", ".hta"},
    "documents": {".doc", ".docx", ".xls", ".xlsx", ".pdf"},
    "archives": {".zip", ".7z", ".rar", ".tar", ".gz"},
}

# Error messages
ERROR_INVALID_OS: Final[str] = "Invalid operating system. Supported: {}"
ERROR_INVALID_VERSION: Final[str] = "Invalid {} version. Supported: {}"
ERROR_INVALID_BITNESS: Final[str] = "Invalid bitness for {}. Supported: {}"
ERROR_INVALID_ENV_TYPE: Final[str] = "Invalid environment type for {}. Supported: {}"
ERROR_INVALID_BROWSER: Final[str] = "Invalid browser for {}. Supported: {}"
ERROR_INVALID_START_FOLDER: Final[str] = "Invalid start folder for {}. Supported: {}"
ERROR_INVALID_PRIVACY: Final[str] = "Invalid privacy type. Supported: {}"
ERROR_INVALID_FILE_EXT: Final[str] = "Invalid file extension. Supported: {}"
ERROR_FILE_TOO_LARGE: Final[str] = "File size exceeds limit of {} bytes"
ERROR_TASK_NOT_FOUND: Final[str] = "Analysis task not found: {}"
ERROR_TASK_COMPLETED: Final[str] = "Analysis task already completed: {}"
ERROR_TASK_FAILED: Final[str] = "Analysis task failed: {}"



"""Version information."""

from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

import tomli


def get_version_from_pyproject() -> str:
    """Get version from pyproject.toml."""
    try:
        pyproject_path = Path(__file__).parent.parent.parent / "pyproject.toml"
        with open(pyproject_path, "rb") as f:
            pyproject = tomli.load(f)
        return f"{pyproject['project']['version']}-dev"
    except Exception:
        return "unknown-dev"


try:
    __version__ = version("anyrun-tools")
except PackageNotFoundError:
    __version__ = get_version_from_pyproject()



"""Analysis models for Sandbox API v1."""

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, model_validator

from .common import HashesApiDto, PrivacyType, ThreatLevelText
from .task_status_update import TaskStatusDto


# Base models
class ObjectType(str, Enum):
    """Analysis object type."""

    FILE = "file"
    URL = "url"
    DOWNLOAD = "download"
    RERUN = "rerun"


class OSType(str, Enum):
    """Operating system type."""

    WINDOWS = "windows"
    LINUX = "linux"


class BitnessType(str, Enum):
    """Operation System bitness type.

    For Linux: only 64
    For Win11: only 64
    For Win7/10: 32 or 64
    """

    X32 = "32"
    X64 = "64"


class WindowsVersion(str, Enum):
    """Windows version."""

    WIN7 = "7"
    WIN10 = "10"
    WIN11 = "11"


class LinuxVersion(str, Enum):
    """Linux version."""

    UBUNTU_22_04_2 = "22.04.2"


class EnvType(str, Enum):
    """Environment type.

    For Windows: clean, office, complete
    For Linux: only office
    """

    CLEAN = "clean"
    OFFICE = "office"
    COMPLETE = "complete"


class Browser(str, Enum):
    """Browser type.

    Default value for Windows: Microsoft Edge
    Allowed values for Windows: Google Chrome, Mozilla Firefox, Internet Explorer, Microsoft Edge
    Default value for Linux: Google Chrome
    Allowed values for Linux: Mozilla Firefox, Google Chrome
    """

    CHROME = "Google Chrome"
    FIREFOX = "Mozilla Firefox"
    IE = "Internet Explorer"
    EDGE = "Microsoft Edge"


class StartFolder(str, Enum):
    """Start folder.

    Linux: desktop, downloads, home, temp
    Windows: appdata, desktop, downloads, home, root, temp, windows
    """

    DESKTOP = "desktop"
    DOWNLOADS = "downloads"
    HOME = "home"
    TEMP = "temp"
    APPDATA = "appdata"  # Windows only
    ROOT = "root"  # Windows only
    WINDOWS = "windows"  # Windows only


class GeoLocation(str, Enum):
    """Geographic location for TOR and residential proxy."""

    FASTEST = "fastest"
    AU = "AU"  # Australia
    BR = "BR"  # Brazil
    CH = "CH"  # Switzerland
    DE = "DE"  # Germany
    FR = "FR"  # France
    GB = "GB"  # United Kingdom
    IT = "IT"  # Italy
    KR = "KR"  # South Korea
    RU = "RU"  # Russia
    US = "US"  # United States


class TaskOptionsDto(BaseModel):
    """Task options model."""

    private: PrivacyType = Field(description="Privacy level")
    whitelist: List[str] = Field(default=[], description="Whitelisted items")
    mitm: bool = Field(description="MITM status")
    fakenet: bool = Field(description="FakeNet status")
    openVPN: str = Field(description="OpenVPN configuration")
    torGeo: Optional[str] = Field(None, description="TOR geography")
    netviator: bool = Field(description="Netviator status")
    netConnected: bool = Field(description="Network connection status")
    network: str = Field(description="Network type")
    logger: str = Field(description="Logger type")
    presentation: bool = Field(description="Presentation mode")
    teamwork: bool = Field(description="Team collaboration")
    reboots: bool = Field(description="System reboots")
    onlyimportant: bool = Field(description="Important events only")
    video: bool = Field(description="Video recording")
    locale: str = Field(description="System locale")
    residentialProxyGeo: Optional[str] = Field(None, description="Residential proxy geography")
    residentialProxy: bool = Field(description="Residential proxy status")
    autoclickerDebugMode: bool = Field(description="Autoclicker debug mode")
    autoclicker: bool = Field(description="Autoclicker status")
    chatGPT: bool = Field(description="ChatGPT integration")


class AnalysisRequest(BaseModel):
    """Analysis request parameters."""

    # Object parameters
    obj_type: ObjectType = Field(default=ObjectType.FILE, description="Type of new task")
    file: Optional[bytes] = Field(None, description="Required when obj_type=file")
    obj_url: Optional[str] = Field(
        None, description="Required when obj_type=url or obj_type=download"
    )
    task_rerun_uuid: Optional[str] = Field(None, description="Required when obj_type=rerun")

    # Environment parameters
    env_os: Optional[OSType] = Field(None, description="Operation System")
    env_version: Optional[str] = Field(None, description="OS version")
    env_bitness: Optional[BitnessType] = Field(None, description="Bitness of Operation System")
    env_type: Optional[EnvType] = Field(None, description="Environment type")
    env_locale: Optional[str] = Field(None, description="Operation system's language")

    # Object execution parameters
    obj_ext_cmd: Optional[str] = Field(
        None,
        min_length=0,
        max_length=256,
        description="Optional command line (Windows only)",
    )
    obj_ext_browser: Optional[Browser] = Field(None, description="Browser type")
    obj_ext_useragent: Optional[str] = Field(
        None,
        min_length=0,
        max_length=256,
        description="User agent for download type",
    )
    obj_ext_elevateprompt: Optional[bool] = Field(None, description="Windows only")
    obj_force_elevation: Optional[bool] = Field(None, description="Windows only")
    auto_confirm_uac: Optional[bool] = Field(None, description="Windows only")
    run_as_root: Optional[bool] = Field(None, description="Linux only")
    obj_ext_extension: Optional[bool] = Field(None, description="Extension enabled")
    obj_ext_startfolder: Optional[StartFolder] = Field(None, description="Start object folder")

    # Network options
    opt_network_connect: Optional[bool] = Field(None, description="Network connection state")
    opt_network_fakenet: Optional[bool] = Field(None, description="FakeNet feature status")
    opt_network_tor: Optional[bool] = Field(None, description="TOR using")
    opt_network_geo: Optional[GeoLocation] = Field(
        None, description="Geographic location for TOR traffic"
    )
    opt_network_mitm: Optional[bool] = Field(None, description="HTTPS MITM proxy option")
    opt_network_residential_proxy: Optional[bool] = Field(
        None, description="Residential proxy for network traffic"
    )
    opt_network_residential_proxy_geo: Optional[GeoLocation] = Field(
        None, description="Geographic location for residential proxy"
    )

    # Timeout options
    opt_timeout: Optional[int] = Field(
        None, ge=10, le=1200, description="Execution time in seconds (10-1200)"
    )

    # Privacy options
    opt_privacy_type: Optional[PrivacyType] = Field(None, description="Privacy settings")
    opt_privacy_hidesource: Optional[bool] = Field(None, description="Option for hiding source URL")

    # Advanced options
    opt_chatgpt: Optional[bool] = Field(None, description="ChatGPT option")
    opt_automated_interactivity: Optional[bool] = Field(
        None, description="Automated Interactivity (ML) option"
    )

    # Tags
    user_tags: Optional[str] = Field(
        None,
        description=(
            "Pattern: a-z, A-Z, 0-9, hyphen (-), comma (,). "
            "Max length per tag: 16 chars, max tags: 8"
        ),
    )

    @model_validator(mode="after")  # type: ignore[misc]
    def validate_required_fields(self) -> "AnalysisRequest":
        """Validate required fields based on obj_type."""
        if self.obj_type == ObjectType.FILE and not self.file:
            raise ValueError("file is required when obj_type is file")
        if self.obj_type in (ObjectType.URL, ObjectType.DOWNLOAD) and not self.obj_url:
            raise ValueError("obj_url is required when obj_type is url or download")
        if self.obj_type == ObjectType.RERUN and not self.task_rerun_uuid:
            raise ValueError("task_rerun_uuid is required when obj_type is rerun")
        return self


class TaskUrlsDto(BaseModel):
    """Task URLs."""

    related: Optional[str] = Field(None, description="URL of the main analysis report")
    json_url: Optional[str] = Field(None, description="URL of the JSON report")
    misp: Optional[str] = Field(None, description="URL of the MISP report")
    pcap: Optional[str] = Field(None, description="URL for downloading PCAP file")
    file: Optional[str] = Field(None, description="URL for downloading the main object")


class BaseTaskHistoryDto(BaseModel):
    """Base class for task history items."""

    uuid: str = Field(
        pattern="^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
        description="Task UUID",
    )
    verdict: ThreatLevelText = Field(description="Analysis verdict")
    date: datetime = Field(description="Analysis creation timestamp")
    tags: List[str] = Field(default=[], description="Analysis tags")
    hashes: HashesApiDto = Field(description="Object hashes")
    name: Optional[str] = Field(None, description="Name of the main object")
    related: Optional[str] = Field(None, description="URL of the main analysis report")
    json_url: Optional[str] = Field(None, description="URL of the JSON report")
    misp: Optional[str] = Field(None, description="URL of the MISP report")
    pcap: Optional[str] = Field(None, description="URL for downloading PCAP file")
    file: Optional[str] = Field(None, description="URL for downloading the main object")


class AnalysisListItem(BaseTaskHistoryDto):
    """Analysis list item model."""

    pass


class TaskHistoryDto(BaseTaskHistoryDto):
    """Task history item."""

    related: str = Field(description="URL of the main analysis report")
    json_url: str = Field(description="URL of the JSON report")
    misp: str = Field(description="URL of the MISP report")
    pcap: Optional[str] = Field(None, description="URL for downloading PCAP file")
    file: Optional[str] = Field(None, description="URL for downloading the main object")


class AnalysisListRequest(BaseModel):
    """Analysis list request parameters."""

    team: bool = Field(default=False, description="Get team history instead of personal")
    skip: int = Field(default=0, ge=0, description="Number of items to skip")
    limit: int = Field(default=25, ge=1, le=100, description="Number of items per page")


class TaskEnvironmentDto(BaseModel):
    """Task environment model."""

    OS: Dict[str, Any] = Field(description="Operating system information")
    software: List[Dict[str, Any]] = Field(default=[], description="Installed software")


class TaskObjectDto(BaseModel):
    """Task object model."""

    names: Dict[str, str] = Field(description="Object names")
    hashes: HashesApiDto = Field(description="Object hashes")
    urls: Optional[Dict[str, str]] = Field(None, description="Object URLs")


class TaskPublicDto(BaseModel):
    """Task public information model."""

    maxAddedTimeReached: bool = Field(
        default=False, description="Maximum allowed task runtime reached"
    )
    objects: TaskObjectDto = Field(description="Task objects information")
    options: TaskOptionsDto = Field(description="Task options")
    environment: TaskEnvironmentDto = Field(description="Task environment information")


class TaskTimesDto(BaseModel):
    """Task timing information."""

    created: datetime = Field(description="Task creation time")
    started: Optional[datetime] = Field(None, description="Task start time")
    completed: Optional[datetime] = Field(None, description="Task completion time")
    addedTime: Optional[datetime] = Field(None, description="Time when additional time was added")


class TaskActionsDto(BaseModel):
    """Task actions model."""

    addTime: bool = Field(default=False, description="Can add time")
    stop: bool = Field(default=False, description="Can stop task")
    delete: bool = Field(default=False, description="Can delete task")


class TaskStatusUpdateDto(BaseModel):
    """Task status update model."""

    task: TaskStatusDto = Field(description="Task status information")
    completed: bool = Field(default=False, description="Task completion status")
    error: Literal[False] = Field(default=False, description="Error status")


class AnalysisData(BaseModel):
    """Analysis data model."""

    taskid: Optional[str] = Field(None, description="Task ID (used in create response)")
    task_id: Optional[str] = Field(None, description="Task ID (used in status response)")
    status: Optional[str] = Field(None, description="Task status")
    completed: Optional[bool] = Field(None, description="Task completion status")
    verdict: Optional[Dict[str, Any]] = Field(None, description="Analysis verdict")
    task: Optional[TaskStatusDto] = Field(None, description="Task status information")

    @model_validator(mode="after")  # type: ignore[misc]
    def convert_taskid(self) -> "AnalysisData":
        """Convert taskid to task_id if needed."""
        if self.taskid and not self.task_id:
            self.task_id = self.taskid
            self.taskid = None
        return self


class AnalysisResponse(BaseModel):
    """Analysis response."""

    error: bool = Field(description="Error flag")
    data: AnalysisData = Field(description="Response data")
    message: Optional[str] = Field(None, description="Error message")

    @model_validator(mode="after")  # type: ignore[misc]
    def validate_data(self) -> "AnalysisResponse":
        """Validate response data."""
        if not self.error and self.data.taskid:
            self.data.task_id = self.data.taskid
            self.data.taskid = None
        return self


class AnalysisListData(BaseModel):
    """Analysis list data model."""

    tasks: List[AnalysisListItem] = Field(description="List of analysis items")


class AnalysisListResponse(BaseModel):
    """Analysis list response."""

    error: bool = Field(description="Error flag")
    data: AnalysisListData = Field(description="Response data")
    message: Optional[str] = Field(None, description="Error message")


class AddTimeResponse(BaseModel):
    """Response model for add time request."""

    error: bool = Field(description="Error flag")
    message: str = Field(description="Response message")


class StopAnalysisResponse(BaseModel):
    """Stop analysis response."""

    error: bool = Field(description="Error flag")
    message: str = Field(description="Response message")


class DeleteAnalysisResponse(BaseModel):
    """Delete analysis response."""

    error: bool = Field(description="Error flag")
    message: str = Field(description="Response message")
    data: Optional[Dict[str, Any]] = Field(None, description="Optional response data")


class AnalysisStatus(str, Enum):
    """Analysis status."""

    QUEUED = "queued"
    STARTING = "starting"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"


class AnalysisResult(BaseModel):
    """Analysis result model."""

    # Analysis information
    uuid: str = Field(description="Analysis UUID")
    permanentUrl: str = Field(description="URL of the main analysis report")
    duration: int = Field(description="Duration of the analysis in seconds")
    creation: int = Field(description="Timestamp of the analysis creation")
    creationText: str = Field(description="Human-readable timestamp of the analysis creation")
    stopExec: Optional[int] = Field(None, description="Timestamp of the analysis completion")
    stopExecText: Optional[str] = Field(
        None, description="Human-readable timestamp of the analysis completion"
    )

    # Reports
    reports: Dict[str, str] = Field(description="URLs of various report formats")
    tags: List[str] = Field(default=[], description="Analysis tags")

    # Sandbox information
    sandbox: Dict[str, Any] = Field(description="Sandbox information")
    options: Dict[str, Any] = Field(description="Analysis options")
    scores: Dict[str, Any] = Field(description="Analysis scores")

    # Content
    content: Dict[str, Any] = Field(description="Analysis content")

    # Environment
    environments: Dict[str, Any] = Field(description="Environment information")
    counters: Dict[str, Any] = Field(description="Analysis counters")

    # Results
    processes: List[Dict[str, Any]] = Field(default=[], description="Process information")
    network: Dict[str, Any] = Field(description="Network activity")
    modified: Dict[str, Any] = Field(description="Modified files and registry")
    incidents: List[Dict[str, Any]] = Field(default=[], description="Detected incidents")
    mitre: List[Dict[str, Any]] = Field(default=[], description="MITRE ATT&CK information")
    malconf: List[Dict[str, Any]] = Field(default=[], description="Malware configuration")
    debugStrings: List[Dict[str, Any]] = Field(default=[], description="Debug strings")

    # Status
    status: str = Field(description="Task completion status")

    # Basic information
    created_at: datetime = Field(description="Creation timestamp")
    started_at: Optional[datetime] = Field(None, description="Start timestamp")
    completed_at: Optional[datetime] = Field(None, description="Completion timestamp")
    error: Optional[str] = Field(None, description="Error message")

    # Object information
    obj_type: ObjectType = Field(description="Object type")
    obj_url: Optional[str] = Field(None, description="Object URL")
    obj_hash: Optional[str] = Field(None, description="Object hash")
    obj_filename: Optional[str] = Field(None, description="Object filename")
    obj_size: Optional[int] = Field(None, description="Object size in bytes")
    obj_mime: Optional[str] = Field(None, description="Object MIME type")

    # Environment information
    env_os: OSType = Field(description="Operating system type")
    env_bitness: BitnessType = Field(description="Operating system bitness")
    env_version: str = Field(description="Operating system version")
    env_type: EnvType = Field(description="Environment type")
    env_browser: Optional[Browser] = Field(None, description="Browser type")

    # Analysis options
    opt_network_connect: bool = Field(description="Network connection state")
    opt_network_fakenet: bool = Field(description="FakeNet feature status")
    opt_network_tor: bool = Field(description="TOR using")
    opt_network_geo: Optional[str] = None
    opt_network_mitm: bool = Field(description="HTTPS MITM proxy option")
    opt_network_residential_proxy: Optional[bool] = None
    opt_network_residential_proxy_geo: Optional[str] = None
    opt_timeout: Optional[int] = None
    opt_privacy_type: PrivacyType = Field(description="Privacy settings")
    opt_privacy_hidesource: bool = Field(description="Option for hiding source URL")
    opt_chatgpt: Optional[bool] = None

    # Analysis results
    result_score: Optional[int] = Field(None, description="Analysis score")
    result_verdict: Optional[str] = Field(None, description="Analysis verdict")
    result_categories: Optional[List[str]] = Field(None, description="Analysis categories")
    result_tags: Optional[List[str]] = Field(None, description="Analysis tags")
    result_mitre: Optional[List[str]] = Field(None, description="MITRE ATT&CK techniques")
    result_iocs: Optional[Dict[str, Any]] = Field(None, description="Extracted IOCs")
    result_files: Optional[Dict[str, Any]] = Field(None, description="Generated files")
    result_screenshots: Optional[Dict[str, Any]] = Field(None, description="Screenshots")
    result_pcap: Optional[Dict[str, Any]] = Field(None, description="Network traffic")
    result_report: Optional[Dict[str, Any]] = Field(None, description="Analysis report")
    result_summary: Optional[Dict[str, Any]] = Field(None, description="Analysis summary")
    result_errors: Optional[List[str]] = Field(None, description="Analysis errors")

    # Additional information
    user_id: Optional[str] = Field(None, description="User ID")
    team_id: Optional[str] = Field(None, description="Team ID")


class FileAnalysisRequest(AnalysisRequest):
    """File analysis request parameters."""

    obj_type: Literal[ObjectType.FILE] = Field(
        default=ObjectType.FILE, frozen=True, description="Type of new task"
    )
    file: bytes = Field(description="File content")


class URLAnalysisRequest(AnalysisRequest):
    """URL analysis request parameters."""

    obj_type: Literal[ObjectType.URL] = Field(
        default=ObjectType.URL, frozen=True, description="Type of new task"
    )
    obj_url: str = Field(description="Target URL")


class DownloadAnalysisRequest(AnalysisRequest):
    """Download analysis request parameters."""

    obj_type: Literal[ObjectType.DOWNLOAD] = Field(
        default=ObjectType.DOWNLOAD, frozen=True, description="Type of new task"
    )
    obj_url: str = Field(description="URL to download and analyze")


class RerunAnalysisRequest(AnalysisRequest):
    """Rerun analysis request parameters."""

    obj_type: Literal[ObjectType.RERUN] = Field(
        default=ObjectType.RERUN, frozen=True, description="Type of new task"
    )
    task_rerun_uuid: str = Field(description="Task UUID to rerun")


# Base response models
class BaseResponseDto(BaseModel):
    """Base response model."""

    error: Literal[False] = Field(False, description="False indicates a successful request")


class ErrorResponseDto(BaseModel):
    """Error response model."""

    error: Literal[True] = Field(True, description="True indicates a failed request")
    message: str = Field(description="Provides information about the error")


class SuccessMessageDto(BaseResponseDto):
    """Success response with message."""

    message: Literal[
        "Add time in task successful",
        "Stop task successful",
        "Delete task successful",
    ] = Field(description="Contains response message")


# Success response models
class TaskIdResponse(BaseModel):
    """Response containing task ID."""

    taskid: str = Field(pattern="^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$")


class TaskHistoryResponse(BaseModel):
    """Response containing task history."""

    tasks: List[Dict[str, Any]]


class UserLimitsResponse(BaseModel):
    """Response containing user limits."""

    limits: Dict[str, Any]


class EnvironmentsResponse(BaseModel):
    """Response containing available environments."""

    environments: List[Dict[str, Any]]


class UserLimitsDto(BaseModel):
    """User account limits."""

    web: Dict[str, int] = Field(description="Defines limits for interactive usage")
    api: Dict[str, int] = Field(description="Defines limits for API usage")
    parallels: Dict[str, int] = Field(description="Defines limits for parallel runs")


class EnvironmentDto(BaseModel):
    """Environment information."""

    os: OSType
    version: Union[WindowsVersion, LinuxVersion]
    bitness: BitnessType
    build: Optional[int] = None
    variant: Optional[str] = None
    type: EnvType
    software: Dict[str, Any]


class SuccessResponseData(BaseModel):
    """Success response data model.

    According to API schema, exactly one of these fields must be present:
    - taskid: for task creation response
    - tasks: for task history response
    - limits: for user limits response
    - environments: for environment info response
    """

    taskid: Optional[str] = Field(
        None, pattern="^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
    )
    tasks: Optional[List[TaskHistoryDto]] = None
    limits: Optional[UserLimitsDto] = None
    environments: Optional[List[EnvironmentDto]] = None

    @model_validator(mode="after")  # type: ignore[misc]
    def validate_exactly_one_field(self) -> "SuccessResponseData":
        """Validate that exactly one field is present."""
        fields = {
            "taskid": self.taskid is not None,
            "tasks": self.tasks is not None,
            "limits": self.limits is not None,
            "environments": self.environments is not None,
        }
        present_fields = [field for field, present in fields.items() if present]
        if len(present_fields) != 1:
            raise ValueError(
                "Exactly one field must be present, "
                f"got {len(present_fields)}: {', '.join(present_fields)}"
            )
        return self


class SuccessResponseDto(BaseResponseDto):
    """Success response model."""

    data: SuccessResponseData = Field(description="Contains response data")


# Detailed data models
class TaskStatusMainObjectDto(BaseModel):
    """Main object information."""

    names: Dict[str, Any] = Field(description="Object names information")
    hashes: HashesApiDto = Field(description="Object hashes")


class TaskStatusVerdictDto(BaseModel):
    """Task verdict information."""

    threat_level: float = Field(description="Threat level of the analysis")
    text: ThreatLevelText = Field(description="Textual description of threat level")


class TaskStatusSpecsDto(BaseModel):
    """Task specifications."""

    autostart: bool = Field(description="Indicates if threat uses autostart feature")
    bad_module_certificate: bool = Field(description="Bad module certificate status")
    bad_process_certificate: bool = Field(description="Bad process certificate status")
    cpu_overrun: bool = Field(description="Indicates if overly high CPU usage was detected")
    crashed_apps: bool = Field(description="Indicates if any application crash was detected")
    crashed_task: bool = Field(description="Indicates if analysis task crash was detected")
    debug_output: bool = Field(description="Indicates if any debug data was extracted")
    executable_dropped: bool = Field(description="Indicates if threat uses dropped executables")
    exploitable: bool = Field(description="Indicates if any known exploit was detected")
    has_trace: bool = Field(description="Has execution trace")
    injects: bool = Field(description="Indicates if threat uses injections")
    known_threat: bool = Field(description="Indicates if known malware was detected")
    low_access: bool = Field(description="Indicates if threat uses low level access")
    malware_config: bool = Field(description="Indicates if malware config was extracted")
    mem_overrun: bool = Field(description="Indicates if overly high RAM usage was detected")
    multiprocessing: bool = Field(description="Indicates if threat uses multiprocessing")
    network_loader: bool = Field(description="Indicates if network download was detected")
    network_threats: bool = Field(description="Indicates if any network threats were detected")
    not_started: bool = Field(description="Indicates if submitted file execution failed")
    process_dump: bool = Field(description="Process memory dump availability")
    reboot: bool = Field(description="System reboot detected")
    service_luncher: bool = Field(description="New service registration detected")
    spam: bool = Field(description="Spam activity detected")
    static_detections: bool = Field(description="Static analysis detections")
    stealing: bool = Field(description="Information stealing detected")
    susp_struct: bool = Field(description="Suspicious structures detected")
    tor: bool = Field(description="TOR usage detected")
    uac_request: bool = Field(description="UAC request detected")


# Report models
class ReportVerdictDto(BaseModel):
    """Report verdict information."""

    score: float = Field(description="Numeric score of the analysis")
    threatLevel: float = Field(description="Threat level of the analysis")
    threatLevelText: ThreatLevelText = Field(description="Textual description of threat level")


class ReportSpecsDto(BaseModel):
    """Report specifications."""

    autoStart: bool = Field(default=False, description="Auto start flag")
    cpuOverrun: bool = Field(default=False, description="CPU overrun flag")
    crashedApps: bool = Field(default=False, description="Crashed apps flag")
    crashedTask: bool = Field(default=False, description="Crashed task flag")
    debugOutput: bool = Field(default=False, description="Debug output flag")
    executableDropped: bool = Field(default=False, description="Executable dropped flag")
    exploitable: bool = Field(default=False, description="Exploitable flag")
    injects: bool = Field(default=False, description="Injects flag")
    knownThreat: bool = Field(default=False, description="Known threat flag")
    lowAccess: bool = Field(default=False, description="Low access flag")
    malwareConfig: bool = Field(default=False, description="Malware config flag")
    memOverrun: bool = Field(default=False, description="Memory overrun flag")
    multiprocessing: bool = Field(default=False, description="Multiprocessing flag")
    networkLoader: bool = Field(default=False, description="Network loader flag")
    networkThreats: bool = Field(default=False, description="Network threats flag")
    notStarted: bool = Field(default=False, description="Not started flag")
    privEscalation: bool = Field(default=False, description="Privilege escalation flag")
    rebooted: bool = Field(default=False, description="Rebooted flag")
    serviceLauncher: bool = Field(default=False, description="Service launcher flag")
    spam: bool = Field(default=False, description="Spam flag")
    staticDetections: bool = Field(default=False, description="Static detections flag")
    stealing: bool = Field(default=False, description="Stealing flag")
    suspStruct: bool = Field(default=False, description="Suspicious structure flag")
    torUsed: bool = Field(default=False, description="TOR used flag")

    class Config:
        """Model configuration."""

        json_schema_extra = {
            "example": {
                "autoStart": False,
                "cpuOverrun": False,
                "crashedApps": False,
                "crashedTask": False,
                "debugOutput": False,
                "executableDropped": False,
                "exploitable": False,
                "injects": False,
                "knownThreat": False,
                "lowAccess": False,
                "malwareConfig": False,
                "memOverrun": False,
                "multiprocessing": False,
                "networkLoader": False,
                "networkThreats": False,
                "notStarted": False,
                "privEscalation": False,
                "rebooted": False,
                "serviceLauncher": False,
                "spam": False,
                "staticDetections": False,
                "stealing": False,
                "suspStruct": False,
                "torUsed": False,
            }
        }

    @model_validator(mode="after")  # type: ignore[misc]
    def validate_fields(self) -> "ReportSpecsDto":
        """Validate that all fields are properly set."""
        return self


class ReportMainObjectDto(BaseModel):
    """Main object information in report."""

    type: ObjectType = Field(description="Type of the main object")
    permanentUrl: str = Field(description="URL for downloading the main object")
    filename: Optional[str] = Field(None, description="Name of the main object file")
    basename: Optional[str] = Field(None, description="Submitted file name/url")
    url: Optional[str] = Field(None, description="Full URL")
    hashes: HashesApiDto = Field(description="Object hashes")
    info: Optional[Dict[str, Any]] = Field(None, description="Additional object information")


class ReportEnvironmentDto(BaseModel):
    """Environment information in report."""

    os: Dict[str, Any] = Field(description="Operating system information")
    software: List[Dict[str, Any]] = Field(description="Installed software information")
    hotfixes: Optional[List[str]] = Field(None, description="Installed hotfixes")


class ReportCountersDto(BaseModel):
    """Counters information in report."""

    processes: Dict[str, int] = Field(description="Process statistics")
    network: Dict[str, int] = Field(description="Network statistics")
    files: Dict[str, int] = Field(description="File statistics")
    registry: Dict[str, int] = Field(description="Registry statistics")
    synchronization: Dict[str, Any] = Field(description="Synchronization statistics")


class TaskReportDto(BaseModel):
    """Complete task report."""

    analysis: Dict[str, Any] = Field(description="Analysis information")
    environments: ReportEnvironmentDto = Field(description="Environment information")
    counters: ReportCountersDto = Field(description="Statistics counters")
    processes: List[Dict[str, Any]] = Field(description="Process information")
    network: Dict[str, Any] = Field(description="Network activity information")
    modified: Dict[str, Any] = Field(description="Modified files and registry")
    incidents: List[Dict[str, Any]] = Field(default=[], description="Detected incidents")
    mitre: List[Dict[str, Any]] = Field(default=[], description="MITRE ATT&CK information")
    malconf: List[Dict[str, Any]] = Field(default=[], description="Malware configuration")
    debugStrings: List[Dict[str, Any]] = Field(default=[], description="Debug strings")
    status: str = Field(description="Task completion status")



"""Common types for ANY.RUN Tools."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel, Field, HttpUrl


class APIVersion(str, Enum):
    """API version."""

    V1 = "v1"
    V2 = "v2"


class APIResponse(BaseModel):
    """Base API response model."""

    error: bool = Field(description="Error flag")
    data: dict[str, Any] = Field(description="Response data")
    message: str | None = Field(None, description="Optional message")


class FileUpload(BaseModel):
    """File upload model."""

    content: str | bytes = Field(description="File content or path")
    filename: str = Field(description="File name")
    content_type: str | None = Field(None, description="Content type")


class APIKey(BaseModel):
    """API key model."""

    key: str = Field(description="API key value")
    name: str | None = Field(None, description="Key name")
    created_at: datetime = Field(description="Creation timestamp")
    expires_at: datetime | None = Field(None, description="Expiration timestamp")
    is_active: bool = Field(default=True, description="Key status")
    permissions: dict[str, bool] = Field(default_factory=dict, description="Key permissions")


class Pagination(BaseModel):
    """Pagination parameters."""

    skip: int = Field(default=0, ge=0, description="Number of items to skip")
    limit: int = Field(default=25, ge=1, le=100, description="Number of items per page")
    total: int | None = Field(None, description="Total number of items")


class WebhookConfig(BaseModel):
    """Webhook configuration."""

    url: HttpUrl = Field(description="Webhook URL")
    secret: str | None = Field(None, description="Webhook secret")
    events: list[str] = Field(description="Event types to subscribe to")
    is_active: bool = Field(default=True, description="Webhook status")
    retry_count: int = Field(default=3, ge=0, le=10, description="Number of retry attempts")
    timeout: int = Field(default=30, ge=1, le=60, description="Request timeout in seconds")


class CacheBackend(str, Enum):
    """Cache backend type."""

    MEMORY = "memory"
    REDIS = "redis"
    NONE = "none"


class RateLimitBackend(str, Enum):
    """Rate limit backend type."""

    MEMORY = "memory"
    REDIS = "redis"
    NONE = "none"


class RetryStrategy(str, Enum):
    """Retry strategy type."""

    EXPONENTIAL = "exponential"
    LINEAR = "linear"
    NONE = "none"


class LogLevel(str, Enum):
    """Log level type."""

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class BaseConfig(BaseModel):
    """Base configuration model."""

    api_key: str = Field(..., description="API key")
    base_url: HttpUrl = Field(..., description="Base URL")
    timeout: float = Field(default=30.0, description="Request timeout in seconds")
    verify_ssl: bool = Field(default=True, description="Verify SSL certificates")
    proxies: dict[str, str] | None = Field(None, description="HTTP/HTTPS proxies")
    user_agent: str | None = Field(None, description="User agent string")
    headers: dict[str, str] = Field(default_factory=dict, description="Additional headers")

    # Cache settings
    cache_enabled: bool = Field(default=True, description="Enable caching")
    cache_backend: CacheBackend = Field(default=CacheBackend.MEMORY, description="Cache backend")
    cache_ttl: int = Field(default=300, description="Cache TTL in seconds")
    cache_prefix: str = Field(default="anyrun:", description="Cache key prefix")

    # Rate limit settings
    rate_limit_enabled: bool = Field(default=True, description="Enable rate limiting")
    rate_limit_backend: RateLimitBackend = Field(
        default=RateLimitBackend.MEMORY, description="Rate limit backend"
    )
    rate_limit: int = Field(default=10, description="Rate limit (requests per second)")
    rate_limit_window: float = Field(default=1.0, description="Rate limit window in seconds")

    # Retry settings
    retry_enabled: bool = Field(default=True, description="Enable retries")
    retry_strategy: RetryStrategy = Field(
        default=RetryStrategy.EXPONENTIAL, description="Retry strategy"
    )
    retry_max_attempts: int = Field(default=3, description="Maximum retry attempts")
    retry_initial_delay: float = Field(default=1.0, description="Initial retry delay in seconds")
    retry_max_delay: float = Field(default=60.0, description="Maximum retry delay in seconds")
    retry_backoff_factor: float = Field(default=2.0, description="Retry backoff factor")

    # Logging settings
    log_level: LogLevel = Field(default=LogLevel.INFO, description="Log level")
    log_format: str = Field(
        default="[{time}] {level} {message}",
        description="Log format string",
    )



"""Request types for Sandbox API v1."""

from enum import Enum


class RequestOSType(str, Enum):
    """Operating system type for requests."""

    WINDOWS = "windows"
    LINUX = "linux"


class RequestNetworkType(str, Enum):
    """Network type for requests."""

    DEFAULT = "default"
    TOR = "tor"
    VPN = "vpn"


class RequestGeoLocation(str, Enum):
    """Geographic location for requests."""

    FASTEST = "fastest"
    AU = "AU"
    BR = "BR"
    CH = "CH"
    DE = "DE"
    FR = "FR"
    GB = "GB"
    IT = "IT"
    KR = "KR"
    RU = "RU"
    US = "US"


class RequestPrivacyType(str, Enum):
    """Privacy type for requests."""

    PUBLIC = "public"
    BYLINK = "bylink"
    OWNER = "owner"
    BYTEAM = "byteam"



"""Sandbox API v1 endpoints."""

# Analysis endpoints
ANALYSIS_CREATE = "/v1/analysis"
ANALYSIS_GET = "/v1/analysis/{task_id}"
ANALYSIS_LIST = "/v1/analysis"
ANALYSIS_MONITOR = "/v1/analysis/monitor/{task_id}"
ANALYSIS_ADD_TIME = "/v1/analysis/addtime/{task_id}"
ANALYSIS_STOP = "/v1/analysis/stop/{task_id}"
ANALYSIS_DELETE = "/v1/analysis/delete/{task_id}"

# Environment endpoints
ENVIRONMENT_INFO = "/v1/environment"

# User endpoints
USER_INFO = "/v1/user"
USER_PRESETS = "/v1/user/presets"



"""Retry mechanism for ANY.RUN Tools."""

from __future__ import annotations

import asyncio
import functools
import random
from dataclasses import dataclass
from typing import Awaitable, Callable, Optional, TypeVar, Union

from loguru import logger
from typing_extensions import ParamSpec

from anyrun.exceptions import APIError, RateLimitError, RetryError, ServerError

T = TypeVar("T")
P = ParamSpec("P")


@dataclass
class RetryConfig:
    """Configuration for retry mechanism."""

    max_attempts: int = 3
    """Maximum number of retry attempts."""

    delay: float = 1.0
    """Base delay between attempts in seconds."""

    max_delay: float = 60.0
    """Maximum delay between attempts in seconds."""

    exponential: bool = True
    """Whether to use exponential backoff."""

    jitter: bool = True
    """Whether to add random jitter to delay."""

    exceptions: Union[type[Exception], list[type[Exception]], None] = None
    """Exception types to retry on. If None, retry on all exceptions."""


def retry(
    max_attempts: int = 3,
    delay: float = 1.0,
    max_delay: float = 60.0,
    exponential: bool = True,
    jitter: bool = True,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Retry decorator for async functions.

    Args:
        max_attempts: Maximum number of attempts
        delay: Initial delay between attempts in seconds
        max_delay: Maximum delay between attempts in seconds
        exponential: Whether to use exponential backoff
        jitter: Whether to add jitter to delay

    Returns:
        Callable: Decorated function
    """

    def decorator(
        func: Callable[P, Awaitable[T]],
    ) -> Callable[P, Awaitable[T]]:
        """Decorator function.

        Args:
            func: Function to decorate

        Returns:
            Callable: Decorated function
        """

        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            """Wrapper function.

            Args:
                *args: Positional arguments
                **kwargs: Keyword arguments

            Returns:
                Any: Function result

            Raises:
                RetryError: If all attempts fail
            """
            # Check if the first argument is a class instance with config attribute
            if args and hasattr(args[0], "config") and hasattr(args[0].config, "retry_enabled"):
                if not args[0].config.retry_enabled:
                    return await func(*args, **kwargs)

            attempt = 1
            current_delay = delay
            last_error: Optional[Union[RateLimitError, APIError]] = None

            while attempt <= max_attempts:
                try:
                    return await func(*args, **kwargs)
                except RateLimitError as exc:
                    last_error = exc
                    if attempt >= max_attempts:
                        break

                    # Use retry_after from RateLimitError
                    current_delay = float(exc.retry_after if exc.retry_after is not None else delay)
                    logger.debug(
                        f"Attempt {attempt} failed with RateLimitError. "
                        f"Retrying in {current_delay:.2f} seconds..."
                    )

                    await asyncio.sleep(current_delay)
                    attempt += 1
                except (ServerError, APIError) as exc:
                    last_error = exc
                    if attempt >= max_attempts:
                        break

                    current_delay = min(
                        current_delay * 2 if exponential else current_delay,
                        max_delay,
                    )

                    if jitter:
                        current_delay *= random.uniform(0.5, 1.5)  # nosec B311

                    logger.debug(
                        f"Attempt {attempt} failed with {type(exc).__name__}. "
                        f"Retrying in {current_delay:.2f} seconds..."
                    )

                    await asyncio.sleep(current_delay)
                    attempt += 1

            if last_error:
                raise last_error
            raise RetryError(attempt, Exception("Unknown error"))

        return wrapper

    return decorator



"""Sandbox API v1 client implementation."""

import json
import os
from typing import Any, AsyncGenerator, Dict, Union, cast

import httpx
from loguru import logger
from pydantic import HttpUrl

from ...exceptions import APIError, ValidationError
from ..base import BaseSandboxClient
from .endpoints import (
    ANALYSIS_ADD_TIME,
    ANALYSIS_CREATE,
    ANALYSIS_DELETE,
    ANALYSIS_GET,
    ANALYSIS_LIST,
    ANALYSIS_MONITOR,
    ANALYSIS_STOP,
    ENVIRONMENT_INFO,
    USER_INFO,
    USER_PRESETS,
)
from .models.analysis import (
    AddTimeResponse,
    AnalysisListRequest,
    AnalysisListResponse,
    AnalysisRequest,
    AnalysisResponse,
    DeleteAnalysisResponse,
    DownloadAnalysisRequest,
    FileAnalysisRequest,
    ObjectType,
    RerunAnalysisRequest,
    StopAnalysisResponse,
    URLAnalysisRequest,
)
from .models.environment import EnvironmentResponse
from .models.task_status_update import TaskStatusUpdateDto
from .models.user import UserInfoResponse, UserPresetsResponse


class SandboxClientV1(
    BaseSandboxClient[AnalysisResponse, AnalysisListResponse, EnvironmentResponse]
):
    """Sandbox API v1 client implementation."""

    def _get_endpoint(self, path: str) -> str:
        """Get versioned API endpoint.

        Args:
            path: API path

        Returns:
            str: Full API endpoint with version
        """
        return path

    def _get_file_content(self, file: Union[str, bytes]) -> bytes:
        """Get file content as bytes.

        Args:
            file: File content as string or bytes

        Returns:
            bytes: File content as bytes

        Raises:
            ValidationError: If file content is invalid
        """
        if isinstance(file, str):
            if os.path.isfile(file):
                with open(file, "rb") as f:
                    return f.read()
            return file.encode()
        return file

    async def _handle_response(self, response: httpx.Response) -> Dict[str, Any]:
        """Handle API response.

        Args:
            response: HTTP response

        Returns:
            Response data as dictionary

        Raises:
            APIError: If response is invalid
        """
        # Handle SSE responses
        if response.headers.get("content-type", "").startswith("text/event-stream"):
            try:
                # Extract data from SSE response
                result = json.loads(response.text.replace("data: ", ""))
                if not isinstance(result, dict):
                    raise APIError("SSE response data is not a dictionary")
                return result
            except json.JSONDecodeError:
                raise APIError(f"Invalid SSE JSON response: {response.text}")

        # Handle regular JSON responses
        try:
            data = response.json()
            if not isinstance(data, dict):
                raise APIError("Response data is not a dictionary")
            return data
        except json.JSONDecodeError:
            raise APIError(f"Invalid JSON response: {response.text}")

    async def analyze(self, **kwargs: Any) -> AnalysisResponse:
        """Submit new analysis.

        Args:
            **kwargs: Analysis parameters (see AnalysisRequest schema)

        Returns:
            AnalysisResponse: Analysis response

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        try:
            request = AnalysisRequest(**kwargs)
            data = request.model_dump(exclude_none=True)  # Only include non-None fields

            # Convert enum values to strings
            for key, value in data.items():
                if hasattr(value, "value"):
                    data[key] = value.value

            logger.debug(f"Analysis request data: {data}")

            files = None
            if request.obj_type == ObjectType.FILE:
                if not request.file:
                    raise ValidationError("file is required for file analysis")
                # Handle file content as bytes
                filename = kwargs.get("filename", "malware.exe")
                file_content = self._get_file_content(request.file)
                files = {
                    "file": (
                        filename,
                        file_content,
                    )
                }

            client = await self._ensure_client()
            response = await client.post(
                self._get_endpoint(ANALYSIS_CREATE),
                headers=self._get_headers(),
                data=data,
                files=files,
            )

            logger.debug(f"API response status: {response.status_code}")
            logger.debug(f"API response headers: {response.headers}")
            logger.debug(f"API response text: {response.text}")

            result = await self._handle_response(response)
            return cast(AnalysisResponse, AnalysisResponse.model_validate(result))

        except ValidationError as e:
            logger.error(f"Validation error: {str(e)}")
            raise ValidationError(f"Invalid analysis parameters: {str(e)}")
        except Exception as e:
            logger.error(f"Analysis error: {str(e)}")
            logger.error(f"Error type: {type(e)}")
            if isinstance(e, httpx.HTTPError):
                logger.error(f"HTTP error details: {str(e)}")
            raise APIError(f"Failed to submit analysis: {str(e)}")

    async def get_analysis(self, task_id: str) -> AnalysisResponse:
        """Get analysis information.

        Args:
            task_id: Analysis task ID

        Returns:
            AnalysisResponse: Analysis information

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.get(
            self._get_endpoint(ANALYSIS_GET.format(task_id=task_id)),
            headers=self._get_headers(),
        )

        result = await self._handle_response(response)
        return cast(AnalysisResponse, AnalysisResponse.model_validate(result))

    async def list_analyses(self, **kwargs: Any) -> AnalysisListResponse:
        """Get list of analyses.

        Args:
            **kwargs: List parameters (see AnalysisListRequest schema)

        Returns:
            AnalysisListResponse: List of analyses

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        try:
            request = AnalysisListRequest(**kwargs)
            client = await self._ensure_client()
            response = await client.get(
                self._get_endpoint(ANALYSIS_LIST),
                headers=self._get_headers(),
                params=request.model_dump(exclude_unset=True),
            )

            result = await self._handle_response(response)
            return cast(AnalysisListResponse, AnalysisListResponse.model_validate(result))

        except ValidationError as e:
            raise ValidationError(f"Invalid list parameters: {str(e)}")
        except Exception as e:
            raise APIError(f"Failed to list analyses: {str(e)}")

    async def get_environment(self) -> EnvironmentResponse:
        """Get available environment information.

        Returns:
            EnvironmentResponse: Environment information

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.get(
            self._get_endpoint(ENVIRONMENT_INFO),
            headers=self._get_headers(),
        )

        result = await self._handle_response(response)
        return cast(EnvironmentResponse, EnvironmentResponse.model_validate(result))

    async def user_info(self) -> UserInfoResponse:
        """Get user information and limits.

        Returns:
            UserInfoResponse: User information and limits

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.get(
            self._get_endpoint(USER_INFO),
            headers=self._get_headers(),
        )

        result = await self._handle_response(response)
        return cast(UserInfoResponse, UserInfoResponse.model_validate(result))

    async def get_user_presets(self) -> UserPresetsResponse:
        """Get user presets.

        Returns:
            UserPresetsResponse: User presets

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.get(
            self._get_endpoint(USER_PRESETS),
            headers=self._get_headers(),
        )

        try:
            # Handle both array and wrapped response formats
            data = response.json()
            return UserPresetsResponse.model_validate(data)
        except Exception as e:
            logger.error(f"Failed to get user presets. Status: {response.status_code}")
            logger.error(f"Response headers: {dict(response.headers)}")
            logger.error(f"Response text: {response.text}")
            raise APIError(f"Failed to get user presets: {str(e)}")

    async def add_analysis_time(self, task_id: str) -> AddTimeResponse:
        """Add time to running analysis.

        Args:
            task_id: Analysis task ID

        Returns:
            AddTimeResponse: Response data

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.patch(
            self._get_endpoint(ANALYSIS_ADD_TIME.format(task_id=task_id)),
            headers=self._get_headers(),
        )

        result = await self._handle_response(response)
        return cast(AddTimeResponse, AddTimeResponse.model_validate(result))

    async def stop_analysis(self, task_id: str) -> StopAnalysisResponse:
        """Stop running analysis.

        Args:
            task_id: Analysis task ID

        Returns:
            StopAnalysisResponse: Response data

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.patch(
            self._get_endpoint(ANALYSIS_STOP.format(task_id=task_id)),
            headers=self._get_headers(),
        )

        result = await self._handle_response(response)
        return cast(StopAnalysisResponse, StopAnalysisResponse.model_validate(result))

    async def delete_analysis(self, task_id: str) -> DeleteAnalysisResponse:
        """Delete analysis.

        Args:
            task_id: Analysis task ID

        Returns:
            DeleteAnalysisResponse: Response data

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.delete(
            self._get_endpoint(ANALYSIS_DELETE.format(task_id=task_id)),
            headers=self._get_headers(),
        )

        result = await self._handle_response(response)
        return cast(DeleteAnalysisResponse, DeleteAnalysisResponse.model_validate(result))

    async def analyze_file(self, file: Union[str, bytes], **kwargs: Any) -> AnalysisResponse:
        """Submit file for analysis.

        Args:
            file: File content as string or bytes
            **kwargs: Additional analysis parameters

        Returns:
            AnalysisResponse: Analysis response

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        kwargs["obj_type"] = ObjectType.FILE
        kwargs["file"] = self._get_file_content(file)
        request = FileAnalysisRequest(**kwargs)
        return await self.analyze(**request.model_dump())

    async def analyze_url(self, url: Union[str, HttpUrl], **kwargs: Any) -> AnalysisResponse:
        """Submit URL for analysis.

        Args:
            url: Target URL
            **kwargs: Additional analysis parameters

        Returns:
            AnalysisResponse: Analysis response

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        kwargs["obj_type"] = ObjectType.URL
        kwargs["obj_url"] = str(url)
        request = URLAnalysisRequest(**kwargs)
        return await self.analyze(**request.model_dump())

    async def rerun_analysis(self, task_uuid: str, **kwargs: Any) -> AnalysisResponse:
        """Rerun existing analysis.

        Args:
            task_uuid: Task UUID to rerun
            **kwargs: Additional analysis parameters

        Returns:
            AnalysisResponse: Analysis response

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        kwargs["obj_type"] = ObjectType.RERUN
        kwargs["task_rerun_uuid"] = task_uuid
        request = RerunAnalysisRequest(**kwargs)
        return await self.analyze(**request.model_dump())

    async def get_analysis_monitor(self, task_id: str) -> Dict[str, Any]:
        """Get analysis monitor data.

        Args:
            task_id: Analysis task ID

        Returns:
            Dict[str, Any]: Monitor data

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        response = await client.get(
            self._get_endpoint(ANALYSIS_MONITOR.format(task_id=task_id)),
            headers=self._get_headers(),
        )

        return await self._handle_response(response)

    async def get_analysis_status_stream(
        self, task_id: str
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """Get analysis status stream.

        Args:
            task_id: Analysis task ID

        Yields:
            Dict[str, Any]: Status update

        Raises:
            APIError: If API request failed
        """
        client = await self._ensure_client()
        async with client.stream(
            "GET",
            self._get_endpoint(ANALYSIS_MONITOR.format(task_id=task_id)),
            headers=self._get_headers(),
        ) as response:
            if response.status_code != 200:
                raise APIError(f"Failed to get analysis status: {response.status_code}")

            async for line in response.aiter_lines():
                if not line.strip():
                    continue
                if not line.startswith("data: "):
                    continue
                try:
                    data = json.loads(line[6:])  # Remove "data: " prefix
                    if isinstance(data, dict):
                        # Convert to model and back to dict to validate and normalize data
                        update = TaskStatusUpdateDto.model_validate(data)
                        yield update.model_dump()
                except json.JSONDecodeError as e:
                    raise APIError(f"Invalid SSE JSON response: {line}") from e
                except Exception as e:
                    raise APIError(f"Error processing SSE response: {str(e)}") from e

    async def analyze_download(self, url: Union[str, HttpUrl], **kwargs: Any) -> AnalysisResponse:
        """Submit URL for download and analysis.

        Args:
            url: URL to download and analyze
            **kwargs: Additional analysis parameters

        Returns:
            AnalysisResponse: Analysis response

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        kwargs["obj_type"] = ObjectType.DOWNLOAD
        kwargs["obj_url"] = str(url)
        request = DownloadAnalysisRequest(**kwargs)
        return await self.analyze(**request.model_dump())



"""TI YARA API client implementation."""

from typing import Any, Dict, Optional

import httpx


class TIYaraClient:
    """TI YARA API client."""

    def __init__(
        self,
        api_key: str,
        version: str = "v1",
        cache_enabled: bool = True,
        timeout: int = 30,
    ) -> None:
        """Initialize TI YARA client.

        Args:
            api_key: ANY.RUN API key
            version: API version
            cache_enabled: Enable caching
            timeout: Request timeout in seconds
        """
        self.api_key = api_key
        self.version = version
        self.cache_enabled = cache_enabled
        self.timeout = timeout
        self.client: Optional[httpx.AsyncClient] = None

    async def match_rule(self, rule_id: str, content: str) -> Dict[str, Any]:
        """Match YARA rule against content.

        Args:
            rule_id: YARA rule ID
            content: Content to match against

        Returns:
            Dict[str, Any]: Match results

        Raises:
            NotImplementedError: This is a stub
        """
        raise NotImplementedError("TI YARA API is not implemented yet")

    async def close(self) -> None:
        """Close HTTP client."""
        if self.client is not None:
            await self.client.aclose()
            self.client = None

    async def __aenter__(self) -> "TIYaraClient":
        """Enter async context."""
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit async context."""
        await self.close()



"""Cache implementation for ANY.RUN Tools."""

import json
import time
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Optional, Protocol, Tuple, TypeVar, Union

from loguru import logger
from redis.asyncio import Redis

if TYPE_CHECKING:
    from redis.asyncio import Redis

T = TypeVar("T")


class CacheBackend(Protocol):
    """Abstract base class for cache backends."""

    @abstractmethod
    async def get(self, key: str) -> Optional[Any]:
        """Get value from cache.

        Args:
            key: Cache key

        Returns:
            Any: Cached value or None if not found
        """

    @abstractmethod
    async def set(self, key: str, value: Any, ttl: Optional[int] = None) -> None:
        """Set value in cache.

        Args:
            key: Cache key
            value: Value to cache
            ttl: Time to live in seconds
        """

    @abstractmethod
    async def delete(self, key: str) -> None:
        """Delete value from cache.

        Args:
            key: Cache key
        """

    @abstractmethod
    async def exists(self, key: str) -> bool:
        """Check if key exists in cache.

        Args:
            key: Cache key

        Returns:
            bool: True if key exists, False otherwise
        """


class Cache:
    """Cache wrapper with backend selection."""

    def __init__(
        self,
        backend: Optional[Union[CacheBackend, Redis]] = None,
        enabled: bool = True,
        prefix: str = "anyrun:",
        default_ttl: int = 300,
    ) -> None:
        """Initialize cache.

        Args:
            backend: Cache backend instance
            enabled: Whether cache is enabled
            prefix: Key prefix
            default_ttl: Default TTL in seconds
        """
        self.enabled = enabled
        self.prefix = prefix
        self.default_ttl = default_ttl

        if not enabled:
            self.backend: CacheBackend = NoCache()
        elif backend is None:
            self.backend = MemoryCache()
        elif isinstance(backend, (RedisCache, MemoryCache, NoCache)):
            self.backend = backend
        elif isinstance(backend, Redis):
            self.backend = RedisCache(backend, prefix, default_ttl)
        else:
            raise ValueError("Invalid cache backend")

    async def get(self, key: str) -> Optional[Any]:
        """Get value from cache.

        Args:
            key: Cache key

        Returns:
            Any: Cached value or None if not found
        """
        if not self.enabled:
            return None
        return await self.backend.get(key)

    async def set(self, key: str, value: Any, ttl: Optional[int] = None) -> None:
        """Set value in cache.

        Args:
            key: Cache key
            value: Value to cache
            ttl: Time to live in seconds
        """
        if not self.enabled:
            return
        await self.backend.set(key, value, ttl or self.default_ttl)

    async def delete(self, key: str) -> None:
        """Delete value from cache.

        Args:
            key: Cache key
        """
        if not self.enabled:
            return
        await self.backend.delete(key)

    async def exists(self, key: str) -> bool:
        """Check if key exists in cache.

        Args:
            key: Cache key

        Returns:
            bool: True if key exists, False otherwise
        """
        if not self.enabled:
            return False
        return await self.backend.exists(key)


class RedisCache(CacheBackend):
    """Redis cache backend implementation."""

    def __init__(
        self,
        client: Redis,
        prefix: str = "anyrun:",
        default_ttl: int = 300,
    ) -> None:
        """Initialize Redis cache.

        Args:
            client: Redis client instance
            prefix: Key prefix
            default_ttl: Default TTL in seconds
        """
        self.client = client
        self.prefix = prefix
        self.default_ttl = default_ttl

    def _get_key(self, key: str) -> str:
        """Get prefixed key.

        Args:
            key: Cache key

        Returns:
            str: Prefixed key
        """
        return f"{self.prefix}{key}"

    async def get(self, key: str) -> Optional[Any]:
        """Get value from Redis cache.

        Args:
            key: Cache key

        Returns:
            Any: Cached value or None if not found
        """
        try:
            value = await self.client.get(self._get_key(key))
            if value is None:
                return None
            return json.loads(value)
        except Exception as e:
            logger.warning(f"Failed to get value from Redis cache: {e}")
            return None

    async def set(self, key: str, value: Any, ttl: Optional[int] = None) -> None:
        """Set value in Redis cache.

        Args:
            key: Cache key
            value: Value to cache
            ttl: Time to live in seconds
        """
        try:
            await self.client.set(
                self._get_key(key),
                json.dumps(value),
                ex=ttl or self.default_ttl,
            )
        except Exception as e:
            logger.warning(f"Failed to set value in Redis cache: {e}")

    async def delete(self, key: str) -> None:
        """Delete value from Redis cache.

        Args:
            key: Cache key
        """
        try:
            await self.client.delete(self._get_key(key))
        except Exception as e:
            logger.warning(f"Failed to delete value from Redis cache: {e}")

    async def exists(self, key: str) -> bool:
        """Check if key exists in Redis cache.

        Args:
            key: Cache key

        Returns:
            bool: True if key exists, False otherwise
        """
        try:
            return bool(await self.client.exists(self._get_key(key)))
        except Exception as e:
            logger.warning(f"Failed to check key existence in Redis cache: {e}")
            return False


class NoCache(CacheBackend):
    """No-op cache backend implementation."""

    async def get(self, key: str) -> None:
        """Always return None.

        Args:
            key: Cache key

        Returns:
            None: Always returns None
        """
        return None

    async def set(self, key: str, value: Any, ttl: Optional[int] = None) -> None:
        """Do nothing.

        Args:
            key: Cache key
            value: Value to cache
            ttl: Time to live in seconds
        """

    async def delete(self, key: str) -> None:
        """Do nothing.

        Args:
            key: Cache key
        """

    async def exists(self, key: str) -> bool:
        """Always return False.

        Args:
            key: Cache key

        Returns:
            bool: Always returns False
        """
        return False


class MemoryCache(CacheBackend):
    """In-memory cache backend implementation."""

    def __init__(self, prefix: str = "anyrun:", default_ttl: int = 300) -> None:
        """Initialize in-memory cache.

        Args:
            prefix: Key prefix
            default_ttl: Default TTL in seconds
        """
        self._cache: Dict[str, Tuple[Any, Optional[float]]] = {}
        self.prefix = prefix
        self.default_ttl = default_ttl

    def _get_key(self, key: str) -> str:
        """Get prefixed key.

        Args:
            key: Cache key

        Returns:
            str: Prefixed key
        """
        return f"{self.prefix}{key}"

    async def get(self, key: str) -> Optional[Any]:
        """Get value from memory cache.

        Args:
            key: Cache key

        Returns:
            Any: Cached value or None if not found
        """
        key = self._get_key(key)
        if key not in self._cache:
            return None

        value, expiry = self._cache[key]
        if expiry is not None and time.time() > expiry:
            del self._cache[key]
            return None

        return value

    async def set(self, key: str, value: Any, ttl: Optional[int] = None) -> None:
        """Set value in memory cache.

        Args:
            key: Cache key
            value: Value to cache
            ttl: Time to live in seconds
        """
        key = self._get_key(key)
        expiry = time.time() + (ttl or self.default_ttl) if ttl is not None else None
        self._cache[key] = (value, expiry)

    async def delete(self, key: str) -> None:
        """Delete value from memory cache.

        Args:
            key: Cache key
        """
        key = self._get_key(key)
        self._cache.pop(key, None)

    async def exists(self, key: str) -> bool:
        """Check if key exists in memory cache.

        Args:
            key: Cache key

        Returns:
            bool: True if key exists and not expired, False otherwise
        """
        key = self._get_key(key)
        if key not in self._cache:
            return False

        _, expiry = self._cache[key]
        if expiry is not None and time.time() > expiry:
            del self._cache[key]
            return False

        return True


def get_cache_key(prefix: str, *args: Any, **kwargs: Any) -> str:
    """Generate a cache key from prefix and arguments.

    Args:
        prefix: Cache key prefix
        *args: Positional arguments
        **kwargs: Keyword arguments

    Returns:
        Generated cache key
    """
    key_parts = [prefix]
    key_parts.extend(str(arg) for arg in args)
    key_parts.extend(f"{k}={v}" for k, v in sorted(kwargs.items()))
    return ":".join(key_parts)


def get_ttl(key: str, default: Optional[int] = None) -> Optional[int]:
    """Get TTL for cache key.

    Args:
        key: Cache key
        default: Default TTL value

    Returns:
        TTL value in seconds or None if not set
    """



"""Configuration for Sandbox API."""

from pydantic import Field

from ..config import BaseConfig
from ..constants import APIVersion


class SandboxConfig(BaseConfig):
    """Configuration for Sandbox API."""

    # API settings
    api_version: APIVersion = Field(default=APIVersion.V1, description="Sandbox API version")

    # Analysis settings
    auto_download_files: bool = Field(
        default=True, description="Automatically download analysis files"
    )
    download_timeout: int = Field(
        default=300, ge=1, le=3600, description="Download timeout in seconds"
    )
    max_file_size: int = Field(
        default=100 * 1024 * 1024,  # 100MB
        ge=1,
        le=500 * 1024 * 1024,  # 500MB
        description="Maximum file size in bytes",
    )

    # Monitoring settings
    monitor_timeout: int = Field(
        default=1800,  # 30 minutes
        ge=1,
        le=7200,  # 2 hours
        description="Analysis monitoring timeout in seconds",
    )
    monitor_interval: float = Field(
        default=1.0,
        ge=0.1,
        le=60.0,
        description="Analysis monitoring interval in seconds",
    )

    # Cache settings
    environment_cache_ttl: int = Field(
        default=3600,  # 1 hour
        ge=1,
        description="Environment info cache TTL in seconds",
    )
    analysis_cache_ttl: int = Field(
        default=60, ge=1, description="Analysis info cache TTL in seconds"  # 1 minute
    )

    # Rate limits (requests per second)
    analyze_rate_limit: float = Field(
        default=0.2,  # 10 per minute
        ge=0.0,
        le=10.0,
        description="Analysis creation rate limit",
    )
    list_rate_limit: float = Field(
        default=1.0,  # 60 per minute
        ge=0.0,
        le=10.0,
        description="Analysis list rate limit",
    )
    status_rate_limit: float = Field(
        default=2.0,  # 120 per minute
        ge=0.0,
        le=10.0,
        description="Analysis status rate limit",
    )
    download_rate_limit: float = Field(
        default=0.5,  # 30 per minute
        ge=0.0,
        le=10.0,
        description="File download rate limit",
    )

    class Config:
        """Pydantic model configuration."""

        validate_assignment = True
        extra = "forbid"



"""User models for Sandbox API v1."""

from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from .request_types import RequestGeoLocation, RequestNetworkType, RequestOSType, RequestPrivacyType
from .response_types import (
    ResponseGeoLocation,
    ResponseNetworkType,
    ResponseOSType,
    ResponsePrivacyType,
)


class UserInfoRequest(BaseModel):
    """User info request parameters."""

    team: bool = Field(default=False, description="Get team info instead of personal")


class UserLimitsDto(BaseModel):
    """User account limits."""

    web: Dict[str, int] = Field(description="Defines limits for interactive usage")
    api: Dict[str, int] = Field(description="Defines limits for API usage")
    parallels: Dict[str, int] = Field(description="Defines limits for parallel runs")


class UserInfoData(BaseModel):
    """User info data model."""

    limits: UserLimitsDto = Field(description="User account limits")


class UserInfoResponse(BaseModel):
    """User info response."""

    error: Literal[False] = Field(False, description="Error flag")
    data: UserInfoData = Field(description="Response data")


class UserPresetRequest(BaseModel):
    """User preset request model."""

    name: str = Field(min_length=1, max_length=64, description="Preset name")

    # Environment settings
    os: RequestOSType = Field(description="Operating system")
    version: str = Field(pattern="^(7|10|11|22\\.04\\.2)$", description="OS version")
    bitness: int = Field(ge=32, le=64, description="OS bitness")
    type: str = Field(pattern="^(clean|office|complete)$", description="Environment type")
    browser: str = Field(
        pattern="^(Google Chrome|Mozilla Firefox|Internet Explorer|Microsoft Edge)$",
        description="Browser",
    )
    locale: str = Field(description="Locale")
    location: str = Field(
        pattern="^(desktop|downloads|home|temp|appdata|root|windows)$",
        description="Start location",
    )

    # Network settings
    net_connected: bool = Field(alias="netConnected", description="Network connected")
    network: RequestNetworkType = Field(description="Network type")
    fakenet: bool = Field(description="FakeNet enabled")
    mitm: bool = Field(description="MITM enabled")
    netviator: bool = Field(description="Netviator enabled")
    vpn: bool = Field(description="VPN enabled")
    open_vpn: str = Field(alias="openVPN", description="OpenVPN configuration")
    tor_geo: RequestGeoLocation = Field(alias="torGeo", description="TOR geography")
    residential_proxy: bool = Field(
        alias="residentialProxy", description="Residential proxy enabled"
    )
    residential_proxy_geo: RequestGeoLocation = Field(
        alias="residentialProxyGeo", description="Residential proxy geography"
    )

    # Additional settings
    timeout: int = Field(ge=10, le=1200, description="Timeout in seconds")
    privacy: RequestPrivacyType = Field(description="Privacy type")
    hide_source: bool = Field(alias="hide_source", description="Hide source")
    extension: bool = Field(description="Extension enabled")
    autoclicker: bool = Field(description="Autoclicker enabled")
    el: bool = Field(description="Elevation prompt")
    no_controls: bool = Field(alias="noControls", description="No controls")


class UserPresetResponse(BaseModel):
    """User preset response model."""

    id: str = Field(alias="_id", description="Preset ID")
    name: str = Field(min_length=1, max_length=64, description="Preset name")
    user_id: str = Field(alias="userId", description="User ID")
    user_plan_name: str = Field(alias="userPlanName", description="User plan name")
    create_time: datetime = Field(alias="createTime", description="Creation time")

    # Environment settings
    os: ResponseOSType = Field(description="Operating system")
    version: str = Field(pattern="^(7|10|11|22\\.04\\.2)$", description="OS version")
    bitness: int = Field(ge=32, le=64, description="OS bitness")
    type: str = Field(pattern="^(clean|office|complete)$", description="Environment type")
    browser: str = Field(
        pattern="^(Google Chrome|Mozilla Firefox|Internet Explorer|Microsoft Edge)$",
        description="Browser",
    )
    locale: str = Field(description="Locale")
    location: str = Field(
        pattern="^(desktop|downloads|home|temp|appdata|root|windows)$",
        description="Start location",
    )

    # Network settings
    net_connected: bool = Field(alias="netConnected", description="Network connected")
    network: ResponseNetworkType = Field(description="Network type")
    fakenet: bool = Field(description="FakeNet enabled")
    mitm: bool = Field(description="MITM enabled")
    netviator: bool = Field(description="Netviator enabled")
    vpn: bool = Field(description="VPN enabled")
    open_vpn: str = Field(alias="openVPN", description="OpenVPN configuration")
    tor_geo: ResponseGeoLocation = Field(alias="torGeo", description="TOR geography")
    residential_proxy: bool = Field(
        alias="residentialProxy", description="Residential proxy enabled"
    )
    residential_proxy_geo: ResponseGeoLocation = Field(
        alias="residentialProxyGeo", description="Residential proxy geography"
    )

    # Additional settings
    timeout: int = Field(ge=10, le=1200, description="Timeout in seconds")
    privacy: ResponsePrivacyType = Field(description="Privacy type")
    hide_source: bool = Field(alias="hide_source", description="Hide source")
    extension: bool = Field(description="Extension enabled")
    autoclicker: bool = Field(description="Autoclicker enabled")
    el: bool = Field(description="Elevation prompt")
    no_controls: bool = Field(alias="noControls", description="No controls")
    expiration_time: str = Field(alias="expirationTime", description="Expiration time")
    expiration_time_selected: bool = Field(
        alias="expirationTimeSelected", description="Expiration time selected"
    )


class UserPresetsResponse(BaseModel):
    """User presets response."""

    error: Literal[False] = Field(False, description="Error flag")
    data: List[UserPresetResponse] = Field(description="List of user presets")

    @classmethod
    def model_validate(
        cls,
        obj: Any,
        *,
        strict: Optional[bool] = None,
        from_attributes: Optional[bool] = None,
        context: Optional[Any] = None,
    ) -> "UserPresetsResponse":
        """Validate and create model from raw data.

        Args:
            obj: Raw data from API
            strict: Whether to enforce strict validation
            from_attributes: Whether to extract data from object attributes
            context: Optional context for validation

        Returns:
            UserPresetsResponse: Validated model
        """
        if isinstance(obj, list):
            # API returns list of presets directly
            return cls(
                error=False,
                data=[UserPresetResponse.model_validate(item) for item in obj],
            )
        elif isinstance(obj, dict):
            # API returns wrapped response
            return cls(
                error=obj["error"],
                data=[UserPresetResponse.model_validate(item) for item in obj["data"]],
            )
        else:
            raise ValueError(f"Invalid response format: {type(obj)}")



"""Common models for Sandbox API v1."""

from enum import Enum

from pydantic import BaseModel, Field


class ThreatLevelText(str, Enum):
    """Threat level text."""

    UNDETECTED = "No threats detected"
    SUSPICIOUS = "Suspicious activity"
    MALICIOUS = "Malicious activity"


class PrivacyType(str, Enum):
    """Privacy type."""

    PUBLIC = "public"
    BYLINK = "bylink"
    OWNER = "owner"
    BYTEAM = "byteam"


class HashesApiDto(BaseModel):
    """Object's hashes."""

    md5: str = Field(pattern="^[0-9a-f]{32}$", description="MD5 hash string")
    sha1: str = Field(pattern="^[0-9a-f]{40}$", description="SHA1 hash string")
    sha256: str = Field(pattern="^[0-9a-f]{64}$", description="SHA256 hash string")
    ssdeep: str = Field(description="SSDEEP hash string")



"""ANY.RUN Tools - Python SDK for ANY.RUN APIs."""

from ._version import __version__
from .client import AnyRunClient
from .config import BaseConfig
from .exceptions import (
    APIError,
    AuthenticationError,
    ConfigurationError,
    NotFoundError,
    RateLimitError,
    RetryError,
    ServerError,
    ValidationError,
)
from .types import CacheBackend, RetryStrategy

__all__ = [
    "__version__",
    "AnyRunClient",
    "BaseConfig",
    "APIError",
    "AuthenticationError",
    "ConfigurationError",
    "NotFoundError",
    "RateLimitError",
    "RetryError",
    "ServerError",
    "ValidationError",
    "CacheBackend",
    "RetryStrategy",
]



"""Response types for Sandbox API v1."""

from enum import Enum


class ResponseOSType(str, Enum):
    """Operating system type from responses."""

    WINDOWS = "Windows"
    LINUX = "Linux"


class ResponseNetworkType(str, Enum):
    """Network type from responses."""

    DEFAULT = "default"
    TOR = "tor"
    VPN = "vpn"
    UNLIMITED = "unlimited"


class ResponseGeoLocation(str, Enum):
    """Geographic location from responses."""

    FASTEST = "fastest"
    EMPTY = ""
    AU = "AU"
    BR = "BR"
    CH = "CH"
    DE = "DE"
    FR = "FR"
    GB = "GB"
    IT = "IT"
    KR = "KR"
    RU = "RU"
    US = "US"


class ResponsePrivacyType(str, Enum):
    """Privacy type from responses."""

    PUBLIC = "public"
    BYLINK = "bylink"
    OWNER = "owner"
    BYTEAM = "byteam"



"""Exceptions for ANY.RUN API client."""

from typing import Any, Dict, Optional


class AnyRunError(Exception):
    """Base exception for ANY.RUN errors."""

    def __init__(
        self,
        message: str,
        status_code: Optional[int] = None,
        details: Optional[Dict[str, Any]] = None,
        cause: Optional[Exception] = None,
    ) -> None:
        """Initialize exception.

        Args:
            message: Error message
            status_code: Optional HTTP status code
            details: Optional error details
            cause: Optional cause of the error
        """
        super().__init__(message)
        self.status_code = status_code
        self.details = details
        if cause is not None:
            self.__cause__ = cause


class APIError(AnyRunError):
    """Base exception for ANY.RUN API errors."""

    def __init__(
        self,
        message: str,
        status_code: Optional[int] = None,
        details: Optional[Dict[str, Any]] = None,
        cause: Optional[Exception] = None,
    ) -> None:
        """Initialize exception.

        Args:
            message: Error message
            status_code: Optional HTTP status code
            details: Optional error details
            cause: Optional cause of the error
        """
        super().__init__(message, status_code, details, cause)


class RetryError(APIError):
    """Error raised when all retry attempts fail."""

    def __init__(self, attempts: int, last_error: Exception) -> None:
        """Initialize retry error.

        Args:
            attempts: Number of attempts made
            last_error: Last error encountered
        """
        self.attempts = attempts
        self.last_error = last_error
        super().__init__(
            f"Failed after {attempts} attempts. Last error: {str(last_error)}", cause=last_error
        )


class AuthenticationError(APIError):
    """Authentication error."""


class ConfigurationError(APIError):
    """Configuration error."""


class NotFoundError(APIError):
    """Resource not found error."""


class RateLimitError(APIError):
    """Rate limit exceeded error."""

    def __init__(
        self,
        message: str,
        status_code: Optional[int] = None,
        retry_after: Optional[int] = 60,  # Default value
        details: Optional[Dict[str, Any]] = None,
        cause: Optional[Exception] = None,
    ) -> None:
        """Initialize exception.

        Args:
            message: Error message
            status_code: Optional HTTP status code
            retry_after: Optional number of seconds to wait before retrying
            details: Optional error details
            cause: Optional cause of the error
        """
        super().__init__(message, status_code, details, cause)
        self.retry_after = retry_after


class ServerError(APIError):
    """Server error."""


class ValidationError(APIError):
    """Validation error."""



"""Configuration for ANY.RUN API client."""

from typing import Dict, TypeVar

from pydantic import BaseModel, ConfigDict, Field, HttpUrl, ValidationInfo, field_validator

from .constants import DEFAULT_TIMEOUT, DEFAULT_USER_AGENT
from .types import CacheBackend, LogLevel, RetryStrategy

T = TypeVar("T", bound="BaseConfig")


class BaseConfig(BaseModel):
    """Base configuration for ANY.RUN client."""

    model_config = ConfigDict(
        validate_assignment=True,
        frozen=False,
        extra="ignore",
    )

    # API configuration
    api_key: str = Field(min_length=1, description="ANY.RUN API key")
    base_url: HttpUrl = Field(description="Base URL for API requests")
    api_version: str = Field(default="v1", description="API version")

    # HTTP client configuration
    timeout: float = Field(default=DEFAULT_TIMEOUT, ge=0, description="Request timeout in seconds")
    user_agent: str = Field(default=DEFAULT_USER_AGENT, description="User agent string")
    verify_ssl: bool = Field(default=True, description="Verify SSL certificates")
    headers: Dict[str, str] = Field(
        default_factory=lambda: {"User-Agent": DEFAULT_USER_AGENT},
        description="Additional HTTP headers",
    )

    @field_validator("headers")  # type: ignore[misc]
    @classmethod
    def merge_headers(cls, v: Dict[str, str], values: ValidationInfo) -> Dict[str, str]:
        """Merge custom headers with default headers.

        Args:
            v: Custom headers
            values: Values being validated

        Returns:
            Dict[str, str]: Merged headers
        """
        result = {"User-Agent": values.data.get("user_agent", DEFAULT_USER_AGENT)}
        result.update(v)
        return result

    # Cache configuration
    cache_enabled: bool = Field(default=True, description="Enable response caching")
    cache_ttl: int = Field(default=300, ge=0, description="Cache TTL in seconds")
    cache_backend: CacheBackend = Field(
        default=CacheBackend.MEMORY,
        description="Cache backend type",
    )

    # Rate limiting configuration
    rate_limit_enabled: bool = Field(default=True, description="Enable rate limiting")

    # Retry configuration
    retry_strategy: RetryStrategy = Field(
        default=RetryStrategy.EXPONENTIAL,
        description="Retry strategy",
    )
    retry_max_attempts: int = Field(default=3, ge=1, description="Maximum retry attempts")
    retry_initial_delay: float = Field(
        default=1.0, ge=0, description="Initial retry delay in seconds"
    )
    retry_max_delay: float = Field(default=60.0, ge=0, description="Maximum retry delay in seconds")

    # Logging settings
    log_level: LogLevel = Field(default=LogLevel.INFO, description="Log level")
    log_format: str = Field(
        default="[{time}] {level} {message}",
        description="Log format string",
    )



"""Sandbox API v1 models."""

from typing import List

from .analysis import AnalysisListResponse, AnalysisResponse
from .environment import EnvironmentResponse
from .user import UserInfoResponse, UserPresetsResponse

__all__: List[str] = [
    "AnalysisResponse",
    "AnalysisListResponse",
    "EnvironmentResponse",
    "UserInfoResponse",
    "UserPresetsResponse",
]



"""Rate limiting implementation for ANY.RUN Tools."""

import asyncio
import time
from dataclasses import dataclass, field
from typing import Dict, TypedDict

from loguru import logger


class RateLimitError(Exception):
    """Error raised when rate limit is exceeded."""

    def __init__(self, retry_after: float) -> None:
        """Initialize rate limit error.

        Args:
            retry_after: Number of seconds to wait before retrying
        """
        self.retry_after = retry_after
        super().__init__(f"Rate limit exceeded. Please wait {retry_after:.1f} seconds.")


class RateLimiterState(TypedDict):
    """Rate limiter state type."""

    tokens: float
    last_update: float
    lock: asyncio.Lock


@dataclass
class RateLimiter:
    """Token bucket rate limiter implementation."""

    rate: float
    """Number of tokens per second."""

    burst: int
    """Maximum number of tokens that can be accumulated."""

    key: str = "default"
    """Rate limiter key for multiple limiters."""

    _state: Dict[str, RateLimiterState] = field(default_factory=dict)
    """Internal state for multiple rate limiters."""

    def __post_init__(self) -> None:
        """Initialize rate limiter state."""
        if self.key not in self._state:
            self._state[self.key] = {
                "tokens": float(self.burst),
                "last_update": time.monotonic(),
                "lock": asyncio.Lock(),
            }

    async def acquire(self, tokens: int = 1, wait: bool = True) -> None:
        """Acquire tokens from the bucket.

        Args:
            tokens: Number of tokens to acquire
            wait: Whether to wait for tokens to become available

        Raises:
            RateLimitError: If tokens are not available and wait is False
            ValueError: If requested tokens exceeds burst size
        """
        if tokens > self.burst:
            raise ValueError(f"Requested tokens ({tokens}) exceeds burst size ({self.burst})")

        state = self._state[self.key]
        async with state["lock"]:
            now = time.monotonic()
            time_passed = now - state["last_update"]

            # Add new tokens based on time passed
            new_tokens = time_passed * self.rate
            state["tokens"] = min(state["tokens"] + new_tokens, float(self.burst))
            state["last_update"] = now

            if state["tokens"] < tokens:
                if not wait:
                    # Calculate time needed for tokens to become available
                    needed = tokens - state["tokens"]
                    retry_after = needed / self.rate
                    raise RateLimitError(retry_after)

                # Wait for tokens to become available
                needed = tokens - state["tokens"]
                wait_time = needed / self.rate
                logger.debug(
                    f"Rate limit reached. Waiting {wait_time:.1f} seconds "
                    f"for {needed:.1f} tokens..."
                )
                await asyncio.sleep(wait_time)

                # Update state after waiting
                state["tokens"] = float(self.burst)
                state["last_update"] = time.monotonic()

            # Consume tokens
            state["tokens"] -= float(tokens)

    async def check(self, tokens: int = 1) -> bool:
        """Check if tokens are available and consume them if available.

        Args:
            tokens: Number of tokens to check

        Returns:
            bool: True if tokens are available and consumed, False otherwise
        """
        try:
            state = self._state[self.key]
            async with state["lock"]:
                now = time.monotonic()
                time_passed = now - state["last_update"]

                # Add new tokens based on time passed
                new_tokens = time_passed * self.rate
                state["tokens"] = min(state["tokens"] + new_tokens, float(self.burst))
                state["last_update"] = now

                # Check if we have enough tokens
                if state["tokens"] < tokens:
                    return False

                # Consume tokens
                state["tokens"] -= float(tokens)
                return True
        except Exception:
            return False

    def get_available_tokens(self) -> float:
        """Get number of available tokens.

        Returns:
            float: Number of available tokens
        """
        try:
            state = self._state[self.key]
            now = time.monotonic()
            time_passed = now - state["last_update"]
            return min(state["tokens"] + time_passed * self.rate, float(self.burst))
        except Exception:
            return 0.0

    def reset(self) -> None:
        """Reset rate limiter state."""
        state = self._state[self.key]
        state["tokens"] = float(self.burst)
        state["last_update"] = time.monotonic()

    def get_state(self) -> dict:
        """Get the current state of the rate limiter.

        Returns:
            dict: The current state of the rate limiter
        """
        return {
            "limit": self.burst,
            "remaining": self.get_available_tokens(),
            "reset": self.reset,
        }



"""Base sandbox client."""

import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Optional, TypeVar

import httpx
from pydantic import HttpUrl

from ..config import BaseConfig
from ..constants import API_BASE_URL, DEFAULT_TIMEOUT, DEFAULT_USER_AGENT
from ..exceptions import APIError, AuthenticationError, NotFoundError, RateLimitError, ServerError

A = TypeVar("A")  # Analysis response type
L = TypeVar("L")  # List response type
E = TypeVar("E")  # Environment response type


class BaseSandboxClient(Generic[A, L, E], ABC):
    """Base class for sandbox API clients."""

    def __init__(
        self,
        api_key: str,
        base_url: Optional[HttpUrl] = None,
        timeout: Optional[int] = None,
        verify_ssl: bool = True,
        user_agent: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        """Initialize sandbox client.

        Args:
            api_key: ANY.RUN API key
            base_url: Base URL for API requests
            timeout: Request timeout in seconds
            verify_ssl: Verify SSL certificates
            user_agent: User agent string
            headers: Additional headers
        """
        self.config = BaseConfig(
            api_key=api_key,
            base_url=base_url or HttpUrl(API_BASE_URL),
            timeout=timeout or DEFAULT_TIMEOUT,
            verify_ssl=verify_ssl,
            user_agent=user_agent or DEFAULT_USER_AGENT,
            headers=headers or {},
        )
        self._client: Optional[httpx.AsyncClient] = None

    async def _ensure_client(self) -> httpx.AsyncClient:
        """Ensure client is initialized.

        Returns:
            httpx.AsyncClient: HTTP client
        """
        if self._client is None:
            self._client = httpx.AsyncClient(
                base_url=str(self.config.base_url),
                timeout=self.config.timeout,
                follow_redirects=True,
                verify=self.config.verify_ssl,
                headers=self.config.headers,
            )
        return self._client

    def _get_headers(self) -> Dict[str, str]:
        """Get request headers with authorization.

        Returns:
            Dict[str, str]: Headers dictionary
        """
        return {"Authorization": f"API-Key {self.config.api_key}"}

    async def _handle_response(self, response: httpx.Response) -> Dict[str, Any]:
        """Handle API response.

        Args:
            response: HTTP response

        Returns:
            Dict[str, Any]: Response data

        Raises:
            AuthenticationError: If authentication failed (401)
            NotFoundError: If resource not found (404)
            RateLimitError: If rate limit exceeded (429)
            ServerError: If server error occurred (5xx)
            APIError: If other API error occurred
        """
        try:
            data = response.json()
            if not isinstance(data, dict):
                raise APIError("Response data is not a dictionary")
        except json.JSONDecodeError:
            raise APIError(f"Invalid JSON response: {response.text}")

        if response.status_code in (200, 201, 202):
            return data
        elif response.status_code == 401:
            raise AuthenticationError("Authentication failed", response.status_code)
        elif response.status_code == 404:
            raise NotFoundError("Resource not found", response.status_code)
        elif response.status_code == 429:
            raise RateLimitError("Rate limit exceeded", response.status_code)
        elif response.status_code >= 500:
            raise ServerError("Server error", response.status_code)
        else:
            raise APIError(f"API error: {response.status_code}", response.status_code)

    async def close(self) -> None:
        """Close HTTP client."""
        if self._client is not None:
            await self._client.aclose()
            self._client = None

    async def __aenter__(self) -> "BaseSandboxClient[A, L, E]":
        """Enter async context."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Exit async context."""
        await self.close()

    @abstractmethod
    async def analyze_file(self, file: bytes, **kwargs: Any) -> A:
        """Submit file for analysis.

        Args:
            file: File content
            **kwargs: Additional analysis parameters

        Returns:
            A: Analysis response

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        pass

    @abstractmethod
    async def get_analysis(self, task_id: str) -> A:
        """Get analysis information.

        Args:
            task_id: Analysis task ID

        Returns:
            A: Analysis information

        Raises:
            APIError: If API request failed
        """
        pass

    @abstractmethod
    async def list_analyses(self, **kwargs: Any) -> L:
        """Get list of analyses.

        Args:
            **kwargs: List parameters

        Returns:
            L: List of analyses

        Raises:
            ValidationError: If parameters are invalid
            APIError: If API request failed
        """
        pass

    @abstractmethod
    async def get_environment(self) -> E:
        """Get available environment information.

        Returns:
            E: Environment information

        Raises:
            APIError: If API request failed
        """
        pass

    @abstractmethod
    async def get_analysis_monitor(self, task_id: str) -> Dict[str, Any]:
        """Get analysis monitor data.

        Args:
            task_id: Analysis task ID

        Returns:
            Dict[str, Any]: Monitor data

        Raises:
            NotImplementedError: If not implemented in subclass
        """
        raise NotImplementedError



"""Environment models for Sandbox API v1."""

from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field


class AppDto(BaseModel):
    """Application information."""

    name: str = Field(description="Application name")
    version: str = Field(description="Application version")


class SoftwareDto(BaseModel):
    """Software information."""

    ie: Dict[str, Any] = Field(default={}, description="Internet Explorer information")
    upps: List[Any] = Field(default=[], description="Updates information")
    apps: List[AppDto] = Field(default=[], description="Installed applications")


class EnvironmentDto(BaseModel):
    """Environment information."""

    os: str = Field(description="Operating system type")
    software: SoftwareDto = Field(description="Installed software")
    bitness: int = Field(description="Environment bitness")
    type: Optional[str] = Field(None, description="Environment type")
    variant: Optional[str] = Field(None, description="Environment variant")
    version: Optional[str] = Field(None, description="Environment version")


class EnvironmentDataDto(BaseModel):
    """Environment data model."""

    environments: List[EnvironmentDto] = Field(description="List of available environments")


class EnvironmentResponse(BaseModel):
    """Environment information response."""

    error: Literal[False] = Field(False, description="Error flag")
    data: EnvironmentDataDto = Field(description="Response data")
